package NTLSystem;

import java.io.*;
import java.util.*;

/**
 * Provides a common starting point for the various NTLSystem
 * programs.  If started as the main class, provides a small shell
 * which allows the loading of the various NTLSystem programs.
 * <P>
 * This class also has static methods which do some of the dirty work
 * for various NTLSystem programs, such as loading properties files,
 * etc.
 * <P>
 * $Id: NTLLauncher.java,v 1.2 1998/07/31 21:05:39 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @version $Revision: 1.2 $
 **/
public class NTLLauncher 
{
    // {{{ Methods

    // {{{ loadNTLSystemProperties
    /**
     * Loads and returns the NTLSystem properties [both default and
     * user-specific] from their default locations.  If the programName
     * argument is given, that program's properties will also be loaded.
     * <P>
     * If the System property <TT>"ntlsys.home"</TT> is not
     * set [settable by a command-line param to the JRE], properties are
     * assumed to be in <TT>"./properties/default.properties"</TT> ...
     * otherwise, they will be looked for in
     * <TT>"{ntlsys.home}/properties/default.properties"</TT>.
     * <P>
     * The System properties [System.getProperties()] will be replaced
     * by the union of the System properties, the user's NTLSystem
     * properties and the application's properties [and their respective
     * defaults].  The following "search path" of Properties is
     * therefore created:
     * <PRE>
     * [ Application-specific user properties ]
     * |
     * +-> [Application-specific default properties ]
     *     |
     *     +-> [ NTLSystem user properties ]
     *         |
     *         +-> [ NTLSystem default properties ]
     * 
     * <P>
     * If the <TT>"ntlsys.home"</TT> value <EM>does</EM> exist, it will
     * be used over whatever is in
     * <TT>"{ntlsys.home}/properties/default.properties"</TT>.
     * <P>
     * See <A HREF="../properties.html">Properties in NTLSystem</A>
     * for more info. about defined Properties and their use.
     *
     * @param     programName The name of a program to which
     *            "-default.properties" and ".properties" will be
     *            appended to find that application's properties
     *            files. If null, application properties are not
     *            loaded [though the default properties are].
     * @return    The Properties collection for this program.
     * @exception FileNotFoundException If the file [contained in the
     *            exception message] cannot be found.
     * @exception IOException If the file named in the exception
     *            message cannot be found.
     **/
    public static Properties loadNTLSystemProperties(String programName)
	throws FileNotFoundException, IOException
    {
	FileInputStream fis;
	Properties defProps = new Properties();
	Properties userProps = new Properties(defProps);
	String baseDir = System.getProperty("ntlsys.ntlhome");
	if (baseDir == null) {
	    baseDir = new String(".");
	}
	String propFN = new String(baseDir +
				   "/properties/default.properties");
	String userFN = new String(baseDir +
				   "/properties/user.properties");
	try {
	    // Load the default properties
	    // If a FileNotFoundException is thrown, let it be thrown up.
	    fis = new FileInputStream(propFN);
	    defProps.load(fis);
	    fis.close();

	    // Attempt to load the user properties
	    try {
		fis = new FileInputStream(userFN);
		userProps.load(fis);
		fis.close();
	    } catch (FileNotFoundException fnfe) {
		// That's fine; this file need not be present.
	    } // end try-catch

	    // Replace the value of ntlsys.home in the properties
	    // collection, if set.
	    if ( System.getProperty("ntlsys.ntlhome") != null) {
		userProps.put("ntlsys.ntlhome",
			      System.getProperty("ntlsys.ntlhome"));
	    }

	    if (programName != null) {
		// If the application exists, attempt to get it's
		// default and user properties.
		Properties appDefProps = new Properties(userProps);
		Properties appProps = new Properties(appDefProps);
		
		String appDefPropFN = baseDir + "/properties/" + programName + "-default.properties";
		String appPropFN = baseDir + "/properties/" + programName + ".properties";

		// If this throws a FileNotFoundException, so be
		// it... let it be thrown up.
		FileInputStream appDefFIS = new FileInputStream(appDefPropFN);
		appDefProps.load(appDefFIS);
	
		try {
		    FileInputStream appFIS = new FileInputStream(appPropFN);
		    appProps.load(appFIS);
		    return appProps;
		} catch (FileNotFoundException fnfe) {
		    // That's fine... this file need not be present.
		    // In which case, return the application default properties
		    return appDefProps;
		} // end inner try-catch
	    } else {
		return userProps;
	    }

	} catch (SecurityException se) {
	    // Shouldn't happen since this is an Application.
	    throw new InternalError("Getting unexpected SecurityException when attempting to load properties." + se.getMessage());
	} // end try-catch
    }
    // }}} loadNTLSystemProperties

    // {{{ main
    public static void main(String args[])
    {
	try {
	    try {
		System.out.println("Default properties:");
		Properties simple = NTLLauncher.loadNTLSystemProperties( null );
		simple.list(new PrintStream(System.out));
	    } catch (FileNotFoundException fnfe) {
		System.out.println(fnfe.getMessage());
		fnfe.printStackTrace();
	    } // end try-catch
	    
	    try {
		System.out.println("VerbLearn properties + Default properties:");
		Properties app = NTLLauncher.loadNTLSystemProperties("ntllauncher");
		app.list(new PrintStream(System.out));
	    } catch (FileNotFoundException fnfe) {
		System.out.println(fnfe.getMessage());
		fnfe.printStackTrace();
	    } // end try-catch
	} catch (IOException ioe) {
	    System.out.println( ioe.getMessage() );
	    ioe.printStackTrace();
	}  // end try-catch
    }
    // }}} main

    // }}} Methods

}
