package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.Random;

/**
 * A discrete probability distribution.  DiscreteDistributions are
 * immutable.
 * <P>
 * $Id: DiscreteDist.java,v 1.19 1998/10/16 01:57:02 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author Josh Sled [jsled@icsi.berkeley.edu]
 * @version $Revision: 1.19 $
 **/
public class DiscreteDist
implements ProbDist
{
  /**
   * The serialization Stream Unique Identifier.  Recompute this
   * [with serialver] when an incompatible change is made.
   */
  static final long serialVersionUID = -2740518966859858066L;
  
  /** The probabilities **/
  protected double[] values;
  /** 
   * The sum of all the probabilities in the DiscreteDist. This is
   * useful to have stored for normalize() and chooseValue().
   **/
  protected double sum;
  /** 
   * A PRNG for creating random values.  This is lazy-created when needed.
   * <P>
   * NOTE: This should be a "weak reference" so it can be gc'd if
   * necessary... 
   **/
  Random prng;

  /**
   * The TextFields used for editing [when editing].
   **/
  private transient JTextField tfs[];
  /**
   * A number formatter used for formatting/parsing values while
   * editing.
   **/
  private transient NumberFormat nf;

  /**
   * No-arg constructor for Externalization/Serialization.
   **/
  public DiscreteDist()
  {
    // Call the super-constructor.
    super();
    values = new double[0];
  }

  /**
   * Constructs an empty DiscreteDist of the specified size.
   *
   * @param     size The size of the distribution.
   **/
  public DiscreteDist(int size)
  {
    super();
    values = new double[size];
  }

  /**
   * Constructs a DiscreteDist from the given set of values.
   *
   * @param     values The initial values for the distribution.
   **/
  public DiscreteDist(double values[])
  {
    super();
    try {
      // For computing the sum.
      double sum = 0.0;
      // size private storage.
      this.values = new double[values.length];
      // copy the values.
      for (int i=0; i<values.length; i++) {
	setProb(i, values[i]);
	sum += values[i];
      } // end for over 0 <= i < values.length
      setSum(sum);
    } catch (ArrayIndexOutOfBoundsException aioobe) {
      // shouldn't happen since the array length is checked.
      throw new InternalError();
    }

    // normalize things so that it's a valid distribution
    normalize();
  }

  /* DISTRIBUTION-SPECIFIC value access methods */

  /**
   * Swaps two values in the distribution.
   *
   * @param     a_idx The first index to swap.
   * @param     b_idx The second index to swap.
   * @exception ArrayIndexOutOfBoundsException If either of the values
   *            are outside the array range.
   **/
  protected void swap(int a_idx, int b_idx)
       throws ArrayIndexOutOfBoundsException
  {
    if ( (a_idx < 0) || (b_idx < 0) ||
	 (a_idx >= values.length) || (b_idx >= values.length) ) {
      throw new ArrayIndexOutOfBoundsException("Values are outside of array size.");
    }
    double tmp = values[a_idx];
    values[a_idx] = values[b_idx];
    values[b_idx] = tmp;
  }


  /**
   * Adds a value to the distribution with a probability of 0.  The
   * user can change it to whatever they want from there.
   **/
  protected void add()
  {
    double oldValues[] = new double[values.length];
    // Copy the old values
    System.arraycopy(values, 0, oldValues, 0, values.length);
    // Create the new array.
    values = new double[oldValues.length + 1];
    // Copy the old values back.
    System.arraycopy(oldValues, 0, values, 0, oldValues.length);
    // Add the new value
    values[values.length-1] = 0.0;
  }

  /**
   * Removes a specific value from the distribution.
   *
   * @param     index The index of the value to remove.  Not error
   *            checked, so be good...
   **/
  protected void remove(int index)
  {
    double oldValues[] = new double[values.length];
    // Copy the old values
    System.arraycopy(values, 0, oldValues, 0, values.length);
    // Create the new array.
    values = new double[oldValues.length - 1];
    // Copy before the removal point.
    for (int i=0; i<index; i++) {
      values[i] = oldValues[i];
    } // end for over i
    // Copy after the removal point.
    for (int j=index+1; j<oldValues.length; j++) {
      values[j-1] = oldValues[j];
    } // end for over j
  }

  /**
   * Gets the sum of the probabilities in the distribution.
   *
   * @return    The sum of the probabilities in the distribution.
   **/
  protected double getSum()
  {
    return sum;
  }

  /**
   * Sets the sum of the probabilities in the distribution.
   *
   * @param     The new sum of the probabilities in the distribution.
   **/
  protected void setSum(double newSum)
  {
    sum = newSum;
  }

  /**
   * Sets the given probability of the distribution to the newValue.
   *
   * @param     index The discrete value of interest.
   * @param     newValue The new value to set it to.
   * @exception ArrayIndexOutOfBoundsException If index is not a valid
   *            index into the DiscreteDist.
   **/
  protected void setProb(int index, double newValue)
       throws ArrayIndexOutOfBoundsException
  {
    double sumDiff = sum - values[index];
    values[index] = newValue;
    // Modify the sum
    sum = sumDiff + newValue;
  }


  /**
   * Gets the probability at a specific index.
   *
   * @param     index The discrete value of interest.
   * @return    A double of the probability of that point in the
   *            distribution.
   **/
  public double getProb(int index)
  {
    double toRet;
    try {
      toRet = values[index];
    } catch (ArrayIndexOutOfBoundsException aioobe) {
      toRet = 0.0;
    } // end try-catch

    return toRet;
  }

  /**
   * Returns the size of the DiscreteDist, ie: the number of discrete
   * points in the distribution.
   * 
   * @return    The size of the DiscreteDist.
   **/
  public int getSize()
  {
    return values.length;
  }

  
  /* Implementations of the ProbDist interface */

  /**
   * Reports on the continuity of the distribution.
   *
   * @return    False, since a discrete dist isn't really continuous.
   **/
  public boolean isContinuous()
  {
    return false;
  }
  

  /**
   * Returns the probability at the given point of interest.
   *
   * @param     fValue The value of interest.
   * @return    The value at the given index if it is specified, 0.0
   *            otherwise. 
   **/
  public double getProb(double fValue)
  {
    double toRet;
    // Check if it's not an integer masquerading as a double.
    // if (fValue - floor(fValue)) != 0, then the fractional part of
    // fValue is > 0, and 0.0 should be returned [since fValue then
    // isn't a valid integer index].
    if ( (fValue - Math.floor(fValue)) != 0.0 ) {
      toRet = 0.0;
    } else {
      toRet = getProb((int)fValue);
    } // end if else
    return toRet;
  }

  /**
   * Randomly choose a value from dist. 
   * <P>
   * If the DiscreteDistribution is { 0.75, 0.05, 0.2 }, '0' will be
   * returned 75% of the time, etc. 
   * 
   * @return    An index of the discrete distribution in accords with
   *            the stored probabilities.
   **/
  public Object chooseValue()
  {
    int i=0;
    Integer wrapper;

    // Lazy-create the PRNG if necessary.
    if (prng == null) {
      prng = new Random(System.currentTimeMillis());
    }
    // Scale a random value to the sum of the distribution.
    // random value will now be in the range [0.0, sum].
    double randVal = prng.nextDouble() * getSum();

    int length = getSize();
    double total = 0.0;
    for (; i<length && randVal>total; i++) {
      total += getProb(i);
    }
    // i will be incremented one last time before the (randVal>total)
    // test fails, so we have an off-by-one error to correct.
    // 
    // For instance: { 0.25, 0.75 }, randVal = 0.1:
    // For i=0, 0<2 && 0.1>0.0 = true, total += 0.25
    // For i=1, 1<2 && 0.1>0.25 = false.
    // Correct index to return is 0, or (i=1)-1...
    wrapper = new Integer(i-1);
    return (Object)wrapper;
  }

  /**
   * Return the index of the maximum likelihood value.
   *
   * @return    The index of the maximum likelihood value of the
   *            distribution. 
   **/
  public Object getMLValue()
  {
    int maxidx = -1;
    double maxVal = -1.0;
    double curVal = -1.0;

    // find the maximum value in the array
    for (int i=0; i<values.length; i++) {
	curVal = getProb( i );
	if (curVal >= maxVal) {
	    maxidx = i;
	    maxVal = curVal;
	}
    }
    return (Object)new Double( maxidx );
  }

    /**
     * Convenience method for calling getMLValue().
     *
     * @return    A properly-casted maximum-likelihood value.
     * @see       #getMLValue()
     **/
    public Double mode()
    { return (Double)getMLValue(); }

    /**
     * Returns the index of the second-maximum likelihood value [the
     * runner-up in VerbLearn parlance.]  This returns null if a
     * second-most-likely value does not exist [meaning the Dist has
     * only one value].
     * <P>
     * <EM>Note</EM> that if a distribution has multiple equal
     * most-likely values [for instance, the distribution is
     * { 0.143, 0.429, 0.429 }], then one of the most likely values
     * will be returned [the former, FWIW].
     *
     * @return    The index of the second-most-likely value
     *            [Double], or null if the dist has only one element.
     * @see       NTLSystem.verblearn.Sense
     **/
    public Object getSMLValue()
    {
	int maxIdx = -1;
	int secondPlace = -1;
	double curVal = -1.0;
	double maxVal = -1.0;

	for (int i=0; i< values.length; i++) {
	    curVal = getProb( i );
	    if (curVal >= maxVal) {
		secondPlace = maxIdx;
		maxIdx = i;
		maxVal = curVal;
	    }
	} // end for over i
	if ( secondPlace == -1 ) {
	    return null;
	} else {
	    return new Double( (double)secondPlace );
	}
    }

    /**
     * Convenience method for calling getSMLValue().
     *
     * @return    The properly-casted second-most-likely value [or
     *            null if the Dist has only one value].
     * @see       #getSMLValue()
     **/
    public Double runnerup()
    { return (Double)getSMLValue(); }

    /**
     * The 'peakedness' of the distribution.  This is defined to be
     * the ratio of the mode [maximum-likelihood] to the runner-up
     * [second-maximum-likelihood].  It ranges from 1.0 [uniform
     * distribution] to infinity [spike distribution].
     *
     * @return    The 'peakedness' of the distribution.
     **/
    public double peakedness()
    {
	Double mode = (Double)getMLValue();
	Double ru = (Double)getSMLValue();
	if ( ru == null || (getProb(ru.doubleValue()) == 0.0) ) {
	    return Double.POSITIVE_INFINITY;
	} else {
	    return (double)((double)getProb(mode.doubleValue()) /
			    (double)getProb(ru.doubleValue())     );
	}
    }

    /**
     * Redefinition of Object.equals(Object o)
     **/
    public boolean equals(Object o)
    {
	// Check for instance equality [and null-value]
	if (! getClass().isInstance(o)) { return false; }
	// Cast for ease of reference
	DiscreteDist dd = (DiscreteDist) o;
	// Check the length of the values.
	if (dd.getSize() != getSize()) { return false; }
	// Otherwise, check all the values themselves.
	boolean areEqual = true;
	int length = getSize();
	// As long as they're equal, keep checking to see if they're not
	// equal.  Stop as soon as we find out they're not equal.
	for (int i=0; i<length && areEqual; i++) {
	    areEqual &= (getProb(i) == dd.getProb(i));
	} // end for over i
    
	// Return their equality status.
	return areEqual;
    }

  /**
   * Normalizes the distribution.
   * <P>
   * This is done by summing terms and setting each term to it's
   * percentage in of the sum... ie:
   * <PRE>
   * value_i = value_i / (value_0 + value_1 + ... + value_n)
   * </PRE>
   **/
  public void normalize()
  {
    // get the sum of terms
    double sum = getSum();
    int length = getSize();
    double tmpD;

    // normalize values
    if (sum != 0.0) {
      // do nothing if there's no values
      for (int j=0; j < length; j++) {
	try {
	  tmpD = (values[j] / sum);
	  setProb(j, tmpD);
	  setSum(1.0);
	} catch (ArrayIndexOutOfBoundsException aioobe) {
	  // shouldn't happen since the array length is checked.
	  throw new InternalError();
	}
      }
    }
  }

  /**
   * Redefinition of Object.toString()
   **/
  public String toString()
  {
    StringBuffer toRet = new StringBuffer(50);
    toRet.append("DiscreteDist{");
    int length = getSize();
    // All but last value, followed by a comma
    for (int i=0; i<length-1; i++) {
      toRet.append(getProb(i));
      toRet.append(", ");
    }
    // Last value without trailing comma
    if (length-1 >= 0) {
      toRet.append(getProb(length-1));
    }
    toRet.append("}");
    return toRet.toString();
  }

  /**
   * Redefinition of Object.clone()
   **/
  public Object clone()
  {
    try {
      DiscreteDist toRet = (DiscreteDist)super.clone();
      int length = getSize();
      toRet.values = new double[length];
      for (int i=0; i<length; i++) {
	toRet.values[i] = this.values[i];
      }
      toRet.sum = this.sum;
      return toRet;
    } catch (CloneNotSupportedException cnse) {
      // shouldn't happen because Cloneable is defined
      throw new InternalError( cnse.getMessage() );
    } catch (ArrayIndexOutOfBoundsException aioobe) {
      // shouldn't happen since the lengths were checked.
      throw new InternalError( aioobe.getMessage() );
    } // end try-catch
  }

  /* Externalizable implementations */

  /**
   * Reads the distribution in from a human-readable, human-editable
   * format as described in <A HREF="writeExternal(ObjectOutput)">
   * writeExternal(...)</A> below.
   *
   * @param     oi The ObjectInput implementer to read from.
   **/
  public void readExternal(ObjectInput oi)
  {
    // FIXME
  }

  /**
   * Writes the distribution in a human-readable, human-editable
   * format.
   * <P>
   * The format is as follows:<BR>
   * FIXME
   *
   * @param     oo The ObjectOutput implementer to write to.
   **/
  public void writeExternal(ObjectOutput oo)
  {
    // FIXME
  }

  public String getDisplayableName()
  {
    return new String("Discrete Dist");
  }

  public Component getEditInplacePanel()
  {
    JPanel toRet = (JPanel)getEditViewPanel(true);
    KeyListener kl = new KeyAdapter() {
      public void keyTyped(KeyEvent ke) {
	try {
	  saveValues();
	} catch (NotBeingEditedException nbee) {
	  // Won't happen.
	} catch (CannotSaveValuesException csve) {
	  // FIXME
	  // What to do, what to do?
	}
      }
    };
    for (int i=0; i<tfs.length; i++) {
      tfs[i].addKeyListener(kl);
    }
    return toRet;
  }

  public Component getEditViewPanel(boolean edit)
  {
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    JPanel toRet = new JPanel(gbl);
    JLabel l;
    JButton b;

    // Create a NumberFormat if necessary
    if (nf == null) {
      nf = NumberFormat.getNumberInstance();
      nf.setMaximumFractionDigits(5);
    }

    // Create a "normalize" button at the top that does the right
    // thing.
    gbc.anchor = GridBagConstraints.CENTER;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    if (edit) {
      b = new JButton("Normalize");
      b.addActionListener( new ActionListener() {
	public void actionPerformed(ActionEvent ae) {
	  DiscreteDist ddTmp = new DiscreteDist(tfs.length);
	  try {
	    for (int i=0; i<tfs.length; i++) {
	      ddTmp.setProb(i, nf.parse(tfs[i].getText()).doubleValue());
	    } // end for over i
	    ddTmp.normalize();
	    for (int j=0; j<tfs.length; j++) {
	      tfs[j].setText( nf.format(new Double(ddTmp.getProb(j))) );
	      tfs[j].setCaretPosition(0);
	    }  // end for over j
	  } catch (ParseException pe) {
	    // What to do, what to do...
	    throw new InternalError(pe.getMessage());
	  }
	} // end actionPerformed
      });
      gbl.setConstraints(b, gbc);
      toRet.add(b);
    }

    // Create the Labels and TextFields
    tfs = new JTextField[values.length];
    for (int i=0; i<values.length; i++) {
      l = new JLabel("Value " + i);
      gbc.fill = GridBagConstraints.NONE;
      gbc.gridwidth = GridBagConstraints.RELATIVE;
      gbl.setConstraints(l, gbc);
      toRet.add(l);

      tfs[i] = new JTextField(nf.format(getProb(i)), 10);
      if (! edit) {
	tfs[i].setEditable(false);
      }
      gbc.fill = GridBagConstraints.HORIZONTAL;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbl.setConstraints(tfs[i], gbc);
      toRet.add(tfs[i]);
    } // end for over i

    return toRet;
  }

  public StringBuffer canSaveValues()
       throws NotBeingEditedException
  {
    StringBuffer toRet = null;
    if (nf == null) { nf = NumberFormat.getNumberInstance(); }
    // Go through the JTextFields and see if any of them cannot be
    // parsed...
    for (int i=0; i<tfs.length; i++) {
      try {
	nf.parse(tfs[i].getText());
      } catch (ParseException pe) {
	if (toRet == null) { toRet = new StringBuffer(100); }
	toRet.append("Cannot parse ");
	toRet.append(tfs[i].getText());
	toRet.append(" as a double.\n");
      } // end try-catch
    } // end for over i
    return toRet;
  }

  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException
  {
    if (tfs == null) {
      throw new NotBeingEditedException("This DiscreteDistribution is not being edited.");
    }
    if (nf == null) { nf = NumberFormat.getNumberInstance(); }
    for (int i=0; i<tfs.length ; i++) {
      try {
	setProb(i, nf.parse(tfs[i].getText()).doubleValue() );
      } catch (ParseException pe) {
	new CannotSaveValuesException(pe.getMessage());
      } // end try-catch
    } // end for over i

  }

  public void doneEditing()
  {
    tfs = null;
    nf = null;
  }

}
