package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.support.*;

/**
 * A Feature-structure, or "Fstruct".  An Fstruct contains a
 * collection of Features, which represent specific data in the
 * system.
 * <P>
 * $Id: Fstruct.java,v 1.22 1998/10/16 01:57:08 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author     Josh Sled [jsled@icsi.berkeley.edu]
 * @author     Tajh Taylor [tajh@icsi.berkeley.edu]
 * @see        ProbDist
 * @see        FeatureEditor
 * @version    $Revision: 1.22 $
 **/
public class Fstruct
implements Serializable, Cloneable, EditableDataElement, EditableDataContainer
{
   /* ATTRIBUTES */

    /**
     * The serialization stream unique identifier.  Recompute [with
     * serialver] when this class changes.
     **/
    static final long serialVersionUID = -9219090405357584877L;

   /** The name of this Fstruct. */
   private String name;

   /** A collection of Features which this Fstruct contains. */
   private StringSortedHashtable features;

   /** The feature dictionary from which features will be taken. */
   private FeatureDictionary dictionary;
  
   /**
    * The properties for this Fstruct.  Properties are meta-data about
    * the Fstruct.
    **/
   private Properties properties;

    /**
     * The getter method.  This is shared across all Fstructs, so
     * there's no sense in using reflection everytime.  This is a
     * cached value. 
     **/
    private static transient Method getter;

    /**
     * The setter method.  This s shared across all Fstructs, so
     * there's no sense in using reflection everytime.  This is a
     * cached value.
     **/
    private static transient Method setter;

   //===========================================
   // locking stuff

  /*
    Until further notice, this is being commented out because
    we are trying a simpler way to implement locking.

    A single lock for write access. 
   private Lock currentWriteLock; 

    A collection of locks for read access. 
   private Hashtable currentReadLocks; 
   */

   //===========================================
  
   //TODO add belief maintenance stuff
   // Commented out until this is finished.
   // private BeliefNet beliefs;
   // Commented out until this is finished.
   // private Vector    beliefUpdates;

   /**
    * The editing frame [when editing].
    **/
   private transient JFrame f;

  /**
   * The FstructViewer object [when editing].
   **/
  private transient FstructViewer fsViewer;
   
   //====================
   /* CONSTRUCTORS */

  
   /**
    * No-arg constructor for basic Fstruct and for
    * Externalization/Serialization. 
    **/
   public Fstruct()
   {
      name = new String("");
      features = new StringSortedHashtable();
      properties = new Properties();
      dictionary = new FeatureDictionary();
      //      currentReadLocks = new Hashtable();
      //      currentWriteLock = null;
   }

   /**
    * Constructs an Fstruct with the given name.
    *
    * @param     name The name of this Fstruct.
    * @param     dictionary The dictionary to use for this Fstruct.
    **/
   public Fstruct(String name, FeatureDictionary fdict)
   {
      this();
      setName(name);
      dictionary = fdict;
   }

   /**
    * Constructs an Fstruct from a string of the form
    *  "f1=v1 f2=v2 f3=v3" and a specified FeatureDictionary
    * TEMPORARILY assume each feature can be in only one fstruct
    *
    * @param     name The name of this Fstruct.
    * @param     dictionary The dictionary to use for this Fstruct.
    * @param     s A string of form "f1=v1 f2=v2 f3=v3"
    **/
   public Fstruct(String name, FeatureDictionary fdict, String s) 
     throws NTLException, NoSuchFeatureException {
       this();
       setName(name);
       if (s == null || s.trim().equals("")) return;
       try {
	 NestedStreamTokenizer input = new 
	   NestedStreamTokenizer(new StringReader(s));
	 int token = input.nextToken();
	 while (token != NestedStreamTokenizer.TT_EOF) {
	   if (token != NestedStreamTokenizer.TT_WORD) {
	     throw(new IOException("expected feature name"));
	   }
	   String featName = input.sval;
	   
	   token = input.nextToken();
	   if (token != '=') {
	     throw(new IOException("expected '='"));
	   }
	   token = input.nextToken();
	   if (token != NestedStreamTokenizer.TT_WORD) {
	     throw(new IOException("expected feature value"));
	   }

	   // NOTE: !!!! This setValue checks type -- here it's not
	   // going to like a string very much. MUST write other method
	   // that will take the input, create a SPIKE PROB. DISTRIBUTION
	   // (here it's discrete, happens to have all prob mass at VALUE)
	   // soon!...
	   this.getFeature(featName).setValue(input.sval);
	   token = input.nextToken();
	 }
       }
       catch (IOException e) {
	 System.err.println("Fstruct-from-string exception: " + e.getMessage());
       }
     }

   /**
    * Construct a Fstruct corresponding to the MAP feature values in 
    * a given Fstruct. (Wrapper method.)
    * <P>
    * FIXME: Where is this called?
    *
    * @param     fs         The original (probabilistic) Fstruct
    **/
    /*
   public Fstruct(Fstruct fs)
     throws NTLException, NoSuchFeatureException {
       this(singletonArray(fs));
     }
    */


   /**
    * Helper to the constructor Fstruct(PFstruct): wrapper.
    *  Trivial case.
    *
    * @param     fs         The original (probabilistic) Fstruct
    **/
   private static Fstruct[] singletonArray(Fstruct fs)
     throws NTLException, NoSuchFeatureException {
       Fstruct[] result = new Fstruct[1];
       result[0] = fs;
       return result;
     }

   /**
    * Construct an Fstruct from an array of Fstructs, resolving
    * conflicts on a feature-by-feature basis by choosing the MAP value
    * from the most peaked distribution across the Fstructs for each
    * feature, so long as the peakedness >= minSetFeature.  Entries
    * in the array may be null.
    * <P>
    * FIXME: Where is this used?
    * <P>
    * This code might fit better in NTLSystem.verblearn.Sense.
    *
    * @param     name The name of this new Fstruct
    * @param     fstructs   An array of (probabilistic) Fstructs
    * @param     minSetFeature The minimum value required to set this feature.
    **/
   public Fstruct(String name,
		  Fstruct[] fstructs,
		  double minSetFeature)
       throws NTLException, NoSuchFeatureException
    {
     this();
     // Loop through all linking features...
     Vector fNames = this.getAllFeatureNames();
     Enumeration fgen = fNames.elements();
     // ...and store the winners in a Vector for an easy, synched
     // 'putFeatures(...)' at the end.
     Vector featsToAdd = new Vector( fNames.size() );
     while (fgen.hasMoreElements()) {
       String featName = (String)fgen.nextElement();
       // Choose value from peakiest distribution:
       int winner = -1;
       double winnerPeakedness = -1;
       for (int i = 0; i < fstructs.length; i++) {
	 if (fstructs[i] != null) {
	   if (fstructs[i].getFeature(featName) != null) {
	     double currentPeakedness =
	       (fstructs[i].getFeature(featName)).peakedness();
	     if (currentPeakedness > winnerPeakedness) {
	       winner = i; // slot #?
	       winnerPeakedness = currentPeakedness;
	     }
	   } else {
	     // nothing right now... 
	   }
	 }
       }
       if (winner != -1) {
	   // FIXME	 fstructs[winner].getFeature(featName).fixDist();
	   // above now put all prob distr in "mode" value, e.g. switch to
	   // spike distribution. Should we not just adopt winning distribution?
	   // featsToAdd.addElement( fstruct );
       }
     } // end while
   }
   
    /**
     * Construct a Fstruct from an array of PFstructs as in the above
     * constructor, except that any features in the Fstruct "override"
     * take precedence.
     * <P>
     * FIXME: Where is this called?
     *
     * @param     fstruct An array of Fstructs, from which the MAP
     *            values are taken on a Feature-by-Feature basis.
     * @param     override An Fstruct who's values override whatever
     *            is chosen from fstructs.
     * @exception NTLException If 
     **/
    /*
    public Fstruct(Fstruct[] fstructs, Fstruct override) 
	throws NTLException, NoSuchFeatureException {
	this(fstructs);
	Enumeration fEnum = (override.getAllFeatureNames()).elements();
	while (fEnum.hasMoreElements()) {
	    String fname = (String)fEnum.nextElement();
	    getFeature(fname).setValue(override.getFeature(fname).getValue());
	}
    }
    */

   /**
    * Sets the name of the Fstruct to the given string.
    *
    * @param     newName The new name of the Fstruct.
    **/
   public void setName(String newName)
   {
      name = newName;
   }


   /**
    * @return    The name of the Fstruct.
    **/
   public String getName()
   {
      return new String(name);
   }


   /**
    * Gets a property from the Fstruct.  The property is indexed by the
    * given name. If the property does not exist, null is returned.
    *
    * @param     propName The name of the requested property.
    * @return    The property's value, or null if the requested
    *            property does not exist.
    **/
   public String getProperty(String propName)
   {
      return (String)properties.get(propName);
   }

   /**
    * Sets a property in the Fstruct.  The property is given by
    * propName, and set to the value propValue.
    *
    * @param     propName The name of the property to set.
    * @param     propValue The value to set the property to.
    **/
   public void setProperty(String propName, String propValue)
   {
      properties.put(propName, propValue);
   }

   /**
    * Un-sets the given property in the Fstruct.
    * 
    * @param     prop The property name to unset.
    **/
   public void unsetProperty(String prop)
   {
      properties.remove(prop);
   }
   /**
    * Returns true if the given property is present in the Fstruct,
    * false otherwise.
    *
    * @param     property The name of the property to check for
    *            existance.
    * @return    True if the property exists in the Fstruct; false
    *            otherwise. 
    **/
   public boolean isProperty(String property)
   {
      return properties.containsKey(property);
   }

   /**
    * Returns an Enumeration of the property names in the properties
    * collection.  The values for these names can then be requested
    * from the <A HREF="#getProperty(String)">getProperty</A> method.
    *
    * @return    An Enumeration of the properties of this Fstruct.
    **/
   public Enumeration getProperties()
   {
      return properties.keys();
   }

  
   //==========================
   // feature access methods
   // these do the actual changing of the f-struct.

   /** Put a collection of feature/value pairs into the f-struct.
       This is the one that ensures consistency via synchronization.
       @param newFeatures An enumeration of Features to be added.
       @exception NTLException If the Enumeration doesn't really
                  contain Features, which is should.
       @exception NoSuchFeatureException If one of the Features in the
                  Enumeration cannot be found in the
		  FeatureDictionary.
       */
   public synchronized void putFeatures(Enumeration newFeatures)
      throws NTLException, NoSuchFeatureException
   {
      FeatureDescriptor oldDescriptor;
      Feature currentFeature;
      Object  retrieved;

      // do this for all the features in the Vector
      while (newFeatures.hasMoreElements())
      {
	 // make sure the feature is actually a feature
	 retrieved = newFeatures.nextElement();
	 if (!(retrieved instanceof Feature))
	    throw new NTLException("Expected Features in " +
				   "enumeration.");

	 currentFeature = (Feature)retrieved;
	 
	 // make sure the feature has a descriptor that is found in the
	 // dictionary, if there is a dictionary
	 /*
	   Commented out 8/5/98: Now that Features can have names
	   independent from their FeatureDescriptors, this check will
	   cause problems... besides, Features contain a pointer to
	   their FeatureDescriptor, so checking that it's in the
	   FeatureDictionary doesn't give us much...

	 if (dictionary != null)
	 {
	    oldDescriptor = currentFeature.getDescriptor();

	    // see if there's any descriptor with this name
	    if (dictionary.isDescriptor(oldDescriptor.getName()))
	    {
	       // see if it's the same exact descriptor object
	       if (!dictionary.isDescriptor(oldDescriptor))
	       {
		  currentFeature.setDescriptor(dictionary.
		     getDescriptor(oldDescriptor.getName()));
	       }
	    } else {
		throw new NoSuchFeatureException("No entry for feature '" +
						 currentFeature.getName() +
						 "' found in dictionary for f-struct '" +
						 getName() + ".");
	    }
	 }
	 */

	 // add it to the feature list
	 features.put(currentFeature.getName(), currentFeature);
	 // add set it's container to be this fstruct.
	 currentFeature.setContainer( this );
      }
   }

   /** Remove a collection of feature/value pairs from the f-struct.
       This is the one that ensures consistency via synchronization.
       @param     newFeatures An enumeration of String names to be
                  removed.
       */
  public synchronized void removeFeatures(Enumeration oldFeatures)
  {
      String  oldFeatureName;
      Object  retrieved;
      
      while (oldFeatures.hasMoreElements())
      {
	 // get a feature name to be removed
	 retrieved = oldFeatures.nextElement();
	 if (retrieved instanceof Feature) {
	   retrieved = ((Feature)retrieved).getName();
	 }
	 oldFeatureName = (String)retrieved;

	 // get the name and remove it from the list
	 if (features.containsKey(oldFeatureName))
	   features.remove(oldFeatureName);
      }
   }

   /** Retrieve a collection of feature/value pairs from the f-struct
       given a list of feature names.
       This is the one that ensures consistency via synchronization.
       @param featureNames  An enumeration of String names to fetch.
       @return A Vector containing all the relevant Features.
       @exception NTLException If the Enumeration doesn't contain
                  String Feature names.
       @exception NoSuchFeatureException If any of the given names in
                  the Enumeration cannot be found in this Fstruct.
       */
   public synchronized Vector getFeatures(Enumeration featureNames)
      throws NTLException, NoSuchFeatureException
   {
      Vector  featureList;
      String  featureName;
      Object  retrieved;
      Feature currentFeature;

      featureList = new Vector();
      while (featureNames.hasMoreElements())
      {
	 // get one feature name
	 retrieved = featureNames.nextElement();
	 if (!(retrieved instanceof String))
	    throw new NTLException("Expected String feature names in " +
				   "Enumeration.");
	 featureName = (String)retrieved;

	 // find the matching feature and add it to the result vector
	 if (!features.containsKey(featureName))
	    throw new NoSuchFeatureException("Feature '" + featureName +
					     "' not found in Fstruct '" +
					     getName() + "'.");
	 currentFeature = (Feature)features.get(featureName);
	 featureList.addElement(currentFeature);
      }

      return featureList;
   }
      
  /** Get a list of all the feature names from an f-struct.
      This is the one that ensures consistency via synchronization.
      @return  Returns a Vector of string names of features.
   */
   public synchronized Vector getAllFeatureNames()
   {
      Vector      names;
      Enumeration hashKeys;

      // set up the array and the enumeration
      names = new Vector( features.size() );
      hashKeys = features.getSortedKeys();

      // pick out the strings and put them in the array
      while ( hashKeys.hasMoreElements() )
      {
	  names.addElement( hashKeys.nextElement() );
      }

      // return the results
      return names;
   }
  
   /** See if a feature by this name is in the fstruct.
      This is the one that ensures consistency via synchronization.
       @param featureName The name of the feature.
       @return A boolean flag indicating whether or not
               the feature was found. */
   public synchronized boolean hasFeature(String featureName)
   {
      return features.containsKey(featureName);
   }

   //============================================================
   // these are just wrapper variations on the above methods
  
   /**
    * Put a feature into the fstruct.  If one exists already, replace it.
    * @param     fName The name of the Feature to put.
    * @param     newFeature The new feature to be added.
    * @exception NTLException If <A HREF="putFeature(String, Object)">
    *            putFeature(...)</A> throws one.
    * @exception NoSuchFeatureException If
    *            <A HREF="putFeature(String, Object)">
    *            putFeature(...)</A> throws one. 
    **/
   public void putFeature(String fName, Feature newFeature)
      throws NTLException, NoSuchFeatureException
   {
      Vector fSet;
      Enumeration fEnum;

      // construct the arguments for putFeatures
      fSet = new Vector();
      fSet.addElement(newFeature);
      fEnum = fSet.elements();

      // run the synched method
      putFeatures(fEnum);
   }

  /** Put a feature into the fstruct.  If one exists already, replace it.
       @param     featureName The name of the new feature.
       @param     newValue    The value of the feature.
       @exception NTLException If this Fstruct doesn't have a valid
                  dictionary. 
       @exception NoSuchFeatureException If the named Feature cannot
                  be found.
       */
   public void putFeature(String featureName, Object newValue)
      throws NTLException, NoSuchFeatureException
   {
      Vector fList;
      Enumeration fEnum;
      Feature newFeature;

      // make sure there is a dictionary
      // otherwise, we can't add this way
      if (dictionary == null) {
	throw new NTLException("Can't find descriptor because there is " +
			       "no dictionary for f-struct '" +
			       getName() + "'.");
      }

      // make sure the name refers to a descriptor that can be
      //  found in the dictionary
      if (!dictionary.isDescriptor(featureName)) {
	throw new NoSuchFeatureException("Descriptor '" + featureName +
					 "' not found in dictionary.");
      }

      // construct the new feature
      newFeature = new Feature(dictionary.getDescriptor(featureName),
			       newValue);
      newFeature.setContainer(this);

      // wrap it in the putFeatures arguments
      fList = new Vector();
      fList.addElement(newFeature);
      fEnum = fList.elements();

      // call the synched method
      putFeatures(fEnum);
   }

  /**
   * Removes a feature from the Fstruct.
   *
   * @param fName     The name of the Feature to remove.
   **/
  public void removeFeature(String fName)
  {
     Vector      fNameList;
     Enumeration fNameEnum;

     // construct the arguments for removeFeatures
     fNameList = new Vector();
     fNameList.addElement(fName);
     fNameEnum = fNameList.elements();

     // run the synched method
     removeFeatures(fNameEnum);
  }

   /** Get a feature from the fstruct based on name.
       @param featureName The name of the desired feature.
       @return            Returns the feature.
       @exception NTLException If <A HREF="#getFeatures(Enumeration)">
                  getFeatures(Enuemration)</A> throws one.
       @exception NoSuchFeatureException If
                  <A HREF="#getFeatures(Enumeration)">
		  getFeatures(Enumeration)</A> throws one.
       */
   public Feature getFeature(String featureName)
      throws NTLException, NoSuchFeatureException
   {
      Vector fNameList;
      Vector fList;
      Enumeration fNameEnum;

      // prepare the arguments for getFeatures
      fNameList = new Vector();
      fNameList.addElement(featureName);
      fNameEnum = fNameList.elements();

      // call the synched method
      fList = getFeatures(fNameEnum);
      return (Feature)fList.firstElement();
   }

   /** Get a handle to the dictionary that we are using.
       @return Returns a handle to a FeatureDictionary.
   */
   public FeatureDictionary getDictionary()
   {
      return dictionary;
   }
     
   
   //=========================================
   /* utility methods */
  
  /** Concatenates the given f-struct onto this one, merging duplicate fields.
      Also copies the corresponding feature dictionary entries, if needed.
      @param otherFstruct the f-struct to be concatenated with this one
      @param doReplace    whether or not the new fields should supercede
                          existing fields
      @exception NTLException If calling <A
                 HREF="#getFeature(String)">getFeature(String)</A>
		 throws one.
  **/
   public void concatenate(Fstruct otherFstruct, boolean doReplace)
      throws NTLException
   {
      FeatureDictionary otherDictionary;
      Vector            otherNames;
      int               i;
      Feature           otherFeature;
      FeatureDescriptor otherDescriptor;
      
      
      // get the list of feature names and the feature dictionary
      // from the other fstruct
      otherDictionary = otherFstruct.getDictionary( );
      otherNames = otherFstruct.getAllFeatureNames( );

      // go through the list of feature names and pick out the
      // corresponding features
      for (i = 0; i < otherNames.size(); i++)
      {
	// get the feature from the other fstruct
	Feature f = otherFstruct.getFeature((String)otherNames.elementAt(i));

	// if we already have this feature *and* doReplace is true,
	// or if we didn't have this feature yet,
	// then we need to put it
	if (((hasFeature((String)otherNames.elementAt(i))) && (doReplace)) ||
	    (!hasFeature((String)otherNames.elementAt(i))))
	  {
	    // put the new dictionary definition first
	    otherDescriptor =
	      otherDictionary.getDescriptor((String)otherNames.elementAt(i));

	    dictionary.putDescriptor(otherDescriptor);

	       // now put the feature name and value
	    putFeature((String)otherNames.elementAt(i), f.getValue());
	  }
      }
   }

    /**
     * Removes all Features whose 'peakedness' is below the given
     * cutoff value.  Features which aren't probabilistic discrete
     * distributions [and thus don't have any notion of peakedness]
     * are ignored. 
     *
     * @param     cutoff The peakedness which Features must be above
     *            to stay in the Fstruct.
     **/
    public synchronized void removeNonPeakedFeatures( double cutoff )
    {
	try {
	    Vector fNames = getAllFeatureNames();
	    Vector features = getFeatures( fNames.elements() );
	    Feature f;
	    DiscreteDist d;
	    for (int i=0; i<features.size(); i++) {
		f = (Feature)features.elementAt( i );
		if ( (f.getDescriptor().isProbabilistic()) &&
		     (f.getValue() instanceof DiscreteDist) ) {
		    if ( ((DiscreteDist)f.getValue()).peakedness() <
			 cutoff ) {
			removeFeature( f.getName() );
		    } // end if ( peakedness < cutoff )
		} // end if (probabilistic &&
  		  //         DiscreteDist value)
	    } // end for over Features
	} catch (NTLException ntle) {
	    // We don't expect NoSuchFeatureException, because we're
	    // asking for what we were just told we have...
	    // We don't expect NTLException, because we're giving it
	    // Strings.
	    throw new InternalError( ntle.getMessage() );
	} // end try-catch
    }


   /* Externalizable implemenation */

   /**
    * Reads the data in the Fstruct in from an ObjectInput
    * implementer.  The data was in a human-(read|edit)able format
    * defined in
    * <A HREF="#writeExternal(ObjectOutput)">writeExternal(...)</A>
    * below.
    *
    * @param     input The ObjectInput implementer to read from.
    **/ 
   public void readExternal(ObjectInput input)
   {
      // FIXME
   }
  

   /**
    * Writes the Fstruct to an ObjectOutput implementer.  The data is
    * in the following human-(read|edit)able format:
    * <P>
    * <PRE>
    * Fstruct {
    *   Name: <name string>
    *   Properties {
    *     <java.util.Properties.save() output>
    *     ... }
    *   Features {
    *     Feature { ... }
    *     Feature { ... }
    *     ... }
    * }
    * </PRE>
    *
    * @param     output The ObjectOutput implementer to write to.
    **/
   public void writeExternal(ObjectOutput output)
   {
      /*
      output.writeObject("Fstruct {");
      output.writeObject("Name: ");
      output.writeObject(name);
      output.writeObject("Properties {");
      output.writeObject(properties);
      output.writeObject("}\nFeatures {");
      output.writeObject(features);
      output.writeObject("}\n}");
      */
   }


   /**
    * Returns a field-by-field copy of the object.  If the component
    * objects are cloneable, their clones will be part of the new
    * object.
    *
    * @return     A clone of the current object.
    **/
   public synchronized Object clone()
   {
      // do the actual cloning;
      try {
	Fstruct toRet = (Fstruct)super.clone();
	toRet.name = new String(getName());
	toRet.features = new StringSortedHashtable();
	Vector newFeatures = new Vector( features.size() );
	Vector fNames = getAllFeatureNames();
	try {
	    Vector feats = getFeatures( fNames.elements() );
	    for (int i=0; i<feats.size(); i++) {
		newFeatures.
		    addElement( (Feature)((Feature)feats.elementAt(i)).clone() );
	    } // end for over i
	    toRet.putFeatures( newFeatures.elements() );
	    toRet.properties = (Properties)properties.clone();
	} catch (NTLException ntle) {
	    // We expect neither:
	    // NTLException [because we know we're passing strings in
	    //   the enumeration]
	    // nor
	    // NoSuchFeatureException [because we're asking for
	    //   features which the Fstruct just told us it has]
	    throw new InternalError( ntle.getMessage() );
	} //end try-catch
	//toRet.beliefs = (BeliefNet)beliefs.clone();
	//toRet.beliefUpdates = (Vector)beliefUpdates.clone();

	return toRet;
      } catch (CloneNotSupportedException cnse) {
	// Won't happen.  super.clone() will succeed...
	throw new InternalError( cnse.getMessage() );
      }
   }

   /**
     Redefinition of Object.equals(Object)
     @param o              The object to be tested.
    **/
   public boolean equals(Object o)
   {
      // Reference equality check.
      if (super.equals(o) == false) { return false; }
      
      // Instance and null check.
      if (getClass().isInstance(o) == false) { return false; }
      Fstruct f = (Fstruct)o;
      
      // Check the name
      if (! f.getName().equals(getName())) { return false; }

      //TODO 
      // Check the properties ?  Necessary?
      // Check the features ?  Necessary?
      return true;
   }

    /**
     * Returns a PropertiesDesc object for this Fstruct object.
     *
     * @return     A PropertiesDesc object which describes this
     *             Fstruct.
     **/
    public PropertiesDesc getPropertiesDesc()
    {
	return new PropertiesDesc( this );
    }

    /**
     * Returns a FeatureDesc object for this Fstruct object, given the
     * specified 'ignore'd and 'irrelevant' Features.
     *
     * @param     ignored The ignored Features.
     * @param     irrelevant The irrelevant Features.
     * @param     showIrrelevant If true, irrelevant Features are
     *            shown.  If false, they are hidden.
     * @return    A FeatureDesc object which describes this Fstruct.
     **/
    public FeatureDesc getFeatureDesc( String[] ignored,
				       String[] irrelevant,
				       boolean showIrrelevant )
    {
	return new FeatureDesc( this,
				ignored, 
				irrelevant,
				showIrrelevant );
    }

   /**
    * Allows for the GUI editing of an Fstruct.  This will bring up a
    * Frame with all the functionality necessary to edit an Fstruct,
    * including:
    * <UL>
    * <LI>Editing the Name of the Fstruct
    * <LI>Adding/Removing/Editing Properties of the Fstruct.
    * <LI>Adding/Remvoing/Editing Features of the Fstruct.
    * </UL>
    *
    * @param     wait If true, the method will return when the editing
    *            window has closed; If false, the method will return
    *            immediately.
    **/
   public void Edit(boolean wait)
   {

     String fsName = getName();
     if (fsName.equals("")) {
       fsName = "<untitled>";
     }

     f = new JFrame("Fstruct Editor - " + fsName);

     GridBagLayout gbl = new GridBagLayout();
     GridBagConstraints c = new GridBagConstraints();

     f.getContentPane().setLayout(gbl);

     JPanel p;
     JButton b;

     // Add the Fstruct viewer.
     c.gridwidth = GridBagConstraints.REMAINDER;
     c.weightx = c.weighty = 1.0;
     c.fill = GridBagConstraints.BOTH;
     Component fsv = getEditPanel();
     gbl.setConstraints(fsv, c);
     f.getContentPane().add(fsv);

     p = new JPanel();
     // Bottom row of one panel containing the Dismiss button.
     c.fill = GridBagConstraints.HORIZONTAL;
     c.weightx = c.weighty = 0.0;
     b = new JButton("Dismiss");
     b.setActionCommand("DISMISS");
     b.addActionListener( new ActionListener() {
       public void actionPerformed(ActionEvent ae) {
	 Vector r = new Vector();
	 try {
	   canSaveValues(r);
	   if (r.size() > 0) {
	     String msgs[] = new String[r.size()];
	     r.copyInto( msgs );
	     JOptionPane.showMessageDialog(null,
					   msgs,
					   "Cannot save Fstruct...",
					   JOptionPane.ERROR_MESSAGE);
	   } else {
	     // Save the values and [if successful] close the window.
	     saveValues();
	     f.dispatchEvent(new WindowEvent(f, WindowEvent.WINDOW_CLOSING));
	   }
	 } catch (NotBeingEditedException nbee) {
	   // Shouldn't happen
	   throw new InternalError(nbee.getMessage());
	 } catch (CannotSaveValuesException csve) {
	   throw new InternalError(csve.getMessage());
	 } // end try-catch
       } // end actionPerformed
     });
     p.add(b);

     gbl.setConstraints(p, c);
     f.getContentPane().add(p);

     // This will automagically add itself as a window listener.
     new WindowCloser(f, false);

     f.pack();
     f.setSize( new Dimension(350, 400) );
     f.setVisible(true);

     // If the parameter "wait" is true, wait until the Editing is done
     // before returning.  The WindowWaiter is perfect for this.
     if (wait) {
       WindowWaiter ww = new WindowWaiter();
       f.addWindowListener(ww);
       ww.WaitUntilDone();
     }
  }

  /**
   * Redefinition of Object.toString().
   **/
  public String toString()
  {
    StringBuffer toRet = new StringBuffer(100);
    toRet.append("Fstruct {\n");
    toRet.append("  Name: "); toRet.append(getName()); toRet.append("\n");
    toRet.append("  Properties {\n");
    StringWriter sw = new StringWriter();
    PrintWriter pw = new PrintWriter(sw);
    properties.list(pw);
    toRet.append(sw.toString());
    toRet.append("  }\n");
    toRet.append("  Features {\n");
    Enumeration e = features.elements();
    for (; e.hasMoreElements(); ) {
      toRet.append( ((Feature)e.nextElement()).toString() );
      toRet.append("\n");
    }
    toRet.append(" }\n");
    toRet.append("}\n");
    return toRet.toString();
  }


    // {{{ EditableDataElement implementation

    public String getKey()
    {
	return getName();
    }

    public Component getEditPanel()
    {
	fsViewer = new FstructViewer( this );
	return fsViewer;
    }

    public Component getViewPanel()
    {
	fsViewer = new FstructViewer( this );
	return fsViewer;
    }

    // }}} EditableDataElement impl.

    /**
     * Returns a Component which can be used to edit the Fstruct,
     * ignoring the given set of features, and deeming the given set
     * of feature irrelevant.
     *
     * @param     ignore The set of Feature names to ignore.
     * @param     irrelevant The set of Feature names to deem
     *            irrelevant. 
     **/
    public Component getExtendedEditor(String[] ignore,
				       String[] irrelevant)
    {
	return (new FstructViewer(this, ignore, irrelevant));
    }
				       

  public void canSaveValues(Vector reasons)
       throws NotBeingEditedException
  {
    if (fsViewer == null) {
      throw new NotBeingEditedException("This Fstruct is not being edited.");
    }
    // No need... Fstruct editing is done in place and errors are
    // handeled locally. 
  }

  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException
  {
    if (fsViewer == null) {
      throw new NotBeingEditedException("This Fstruct is not being edited.");
    }
    // Tell the FstructViewer to save values...
    // FIXME: It seems the new Fstruct viewer edits in place... if this
    // is true, no explicit saveValues call is needed.
    //fsViewer.saveValues();
  }

  public void doneEditing()
  {
    fsViewer = null;
  }

  /* EditableDataContainer implementation */

  public void addElement(String name)
  {
    try {
      putFeature(name, dictionary.getFeature(name));
    } catch (NTLException ntle) {
      // Shouldn't happen
      throw new InternalError(ntle.getMessage());
    }
  }

  public void removeElement(String name)
  {
    removeFeature(name);
  }

  public Enumeration getExistingElements()
  {
    return features.keys();
  }

  /* 
   * Inner classes for an Fstruct.
   */

    /**
     * A class which implements PropertyCollection and returns the
     * editable Properties for this Fstruct.
     * <P>
     * $Id: Fstruct.java,v 1.22 1998/10/16 01:57:08 jsled Exp $<BR>
     * Last editor: $Author: jsled $
     *
     * @author   Joshua Sled [jsled@icsi.berkeley.edu]
     * @see      NTLSystem.support.PropertyCollection
     **/
    protected class PropertiesDesc
	implements PropertyCollection
    {
	private Fstruct fstruct;

	/**
	 * Creates a PropertyCollection-implementing object which
	 * returns the collection of Properties in the Fstruct which can
	 * be edited.  This is <EM>all</EM> Properties in the Fstruct.
	 *
	 * @param     target The Fstruct to return information about.
	 **/
	public PropertiesDesc(Fstruct target)
	{
	    fstruct = target;
	}

	public String getLabel()
	{
	    return (fstruct.getName() + " Properties");
	}
      
	public EditablePropertyDescriptor[] getEditableProperties()
	{
	    try {
		// Add one for the "name" field, which is in this panel.
		EditablePropertyDescriptor toRet[] =
		    new EditablePropertyDescriptor[properties.size()];
		Enumeration e = properties.keys();
		Class stringClass = Class.forName("java.lang.String");
		Class getPropParms[] = { stringClass };
		Class setPropParms[] = { stringClass, stringClass };
		Method getter = fstruct.getClass().getMethod("getProperty",
							     getPropParms);
		Method setter = fstruct.getClass().getMethod("setProperty",
							     setPropParms);
		for (int i=0; (i<toRet.length) && e.hasMoreElements(); i++) {
		    String propName = (String)e.nextElement();
		    toRet[i] =
			new EditablePropertyDescriptor(new String(propName),
						       stringClass,
						       getter,
						       setter,
						       new String(propName),
						       fstruct);
		} // end for over i
		return toRet;
	    } catch (ClassNotFoundException cnfe) {
		// This shouldn't occur.
		throw new InternalError( cnfe.getMessage() );
	    } catch (NoSuchMethodException nsme) {
		throw new InternalError( nsme.getMessage() );
	    } // end try-catch
	}
    } // end class PropertiesDesc

  /**
   * A class which implements PropertyCollection and returns the
   * editable Features for this Fstruct.
   * <P>
   * Features can be deemed "ignored", in which case they won't be
   * displayed by the Fstruct editor.  Features can also be ddemed
   * "irrelevant", in which case they will be displayed, but will be
   * greyed out in some manner.  These states are controlled by the
   * use of Fstruct properties: <EM>&lt;feature name&gt;.ignored</EM>
   * for ignored Features, and <EM>&lt;feature
   * name&gt;.irrelevant</EM> for irrelevant Features.
   *
   * @author   Joshua Sled [jsled@icsi.berkeley.edu]
   * @see      NTLSystem.support.PropertyCollection
   **/
    protected class FeatureDesc
	implements PropertyCollection
    {
	/**
	  * A Vector of ignored Features.  These Features are not to be
	  * displayed in the Editing framework, and yet still exist in
	  * the Fstruct. 
	  **/
	private Vector ignored;
	 
	/**
	  * The target Fstruct which we're returning information
	  * about.
	  **/
	private Fstruct fstruct;

	/**
	 * The status of showing irrelevant Features.
	 **/
	private boolean showIrrel;

	/**
	 * Creates a PropertyCollection of the Features from this
	 * Fstruct.  This version of the constructor doesn't ignore any
	 * Features unless the appropriate property [ &lt;feature
	 * name&gt;.ignore ] is set; irrelevant Features will still be
	 * handled if the '&lt;feature name&gt;.irrelevant' property is
	 * set in the Fstruct. 
	 *
	 * @param     target The Fstruct to return information about.
	 * @param     showIrrelevant If true, irrelevant Features will
	 *            be displayed .  If false, irrelevant Features
	 *            will not be displayed [they will be treated as
	 *            'ignored' Features].
	 **/
	public FeatureDesc(Fstruct target, boolean showIrrelevant) {
	    this(target, null, null, showIrrelevant);
	}

	/**
	 * Creates the FeatureDescriptor returner, having it ignore a
	 * given set of Features, and setting a given set of Features
	 * to be deemed 'irrelevant', and are to be greyed out.
	 *
	 * @param     target The Fstruct to return information about.
	 * @param     ignoredFeatures An array of Feature names which
	 *            should not be displayed in editing.
	 * @param     irrelevantFeatures An array of Feature names which
	 *            should be marked as "irrelevant" during editing.
	 * @param     showIrrelevant If true, irrelevant Features will
	 *            be displayed .  If false, irrelevant Features
	 *            will not be displayed [they will be treated as
	 *            'ignored' Features].
	 **/
	public FeatureDesc(Fstruct target,
			   String[] ignoredFeatures,
			   String[] irrelevantFeatures,
			   boolean showIrrelevant)
	{
	    fstruct = target;
	    showIrrel = showIrrelevant;

	    // Save the ignored features for use in getEditableProperties()
	    ignored = new Vector();
	    if (ignoredFeatures != null) {
		ignored.ensureCapacity( ignoredFeatures.length );
		for (int i=0 ; i<ignoredFeatures.length; i++) {
		    ignored.addElement( ignoredFeatures[i] );
		} // end for over i
	    } // end if

	    // For each irrelevant feature, add the "<fname>.irrelevant"
	    // property.  The FeatureEditor class will handle
	    // greying-out these Features.
	    if (irrelevantFeatures != null) {
		for (int j=0; j<irrelevantFeatures.length; j++) {
		    setProperty( irrelevantFeatures + ".irrelevant", "" );
		} // end for over j
	    } // end if
	}

	/**
	  * Sets the state of showing irrelevant Features.
	  * 
	  * @param     newState The new state of showing irrelevant
	  *            Features.  If true, irrelevant Features will be
	  *            shown.  If false, irrelevant Features will not
	  *            be shown [like 'ignored' Features].
	  **/
	public void setIrrelevantFeatureShowing(boolean newState)
	{
	    showIrrel = newState;
	}

	public String getLabel()
	{
	    return (fstruct.getName() + " Features");
	}

	public EditablePropertyDescriptor[] getEditableProperties()
	{
	    Feature f;
	    String name;
	    Vector feats;
	    try {
		feats =
		    fstruct.getFeatures( fstruct.getAllFeatureNames().elements() );
	    } catch (NoSuchFeatureException nsfe) {
		// Shouldn't happen, as we're just feeding back
		// names that we've been given, which _should_ be
		// in the Fstruct...
		throw new InternalError( nsfe.getMessage() );
	    } catch (NTLException ntle) {
		// Shouldn't happen, since we _are_ giving it a
		// String.
		throw new InternalError( ntle.getMessage() );
	    } // end try-catch

	    Vector epds = new Vector( feats.size() );

	    if ( (getter == null) && (setter == null) ) {
		try {
		    Class getterParms[] = { Class.forName("java.lang.String") };
		    Class setterParms[] = { Class.forName("java.lang.String"),
					    Class.forName("NTLSystem.features.Feature") };
		    getter = fstruct.getClass().getMethod("getFeature", getterParms);
		    setter = fstruct.getClass().getMethod("putFeature", setterParms);
		} catch (ClassNotFoundException cnfe) {
		    // Shouldn't happen
		    throw new InternalError(cnfe.getMessage());
		} catch (NoSuchMethodException nsme) {
		    // Shouldn't happen
		    throw new InternalError(nsme.getMessage());
		}
	    } // end if

	    for (int i=0; i < feats.size(); i++) {
		f = (Feature)feats.elementAt( i );
		name = f.getName();
		// If the Feature is ignored or (irrelevant and we're not
		// showing irrelevant Features) then don't display it.
		if ( ( (! ignored.contains(name)) &&
		       (fstruct.getProperty(name + ".ignore") == null) ) &&
		     !((fstruct.getProperty(name + ".irrelevant") != null) &&
		       (!showIrrel)) ) {

		    if (fstruct.getProperty( name + ".irrelevant") != null) {
			name = name + "\nIrrelevant";
		    }
		    epds.addElement( new EditablePropertyDescriptor(name,
								    f.getClass(),
								    getter,
								    setter,
								    f.getName(),
								    fstruct) );
		} // end if
	    } // end for over e

	    EditablePropertyDescriptor toRet[] =
		new EditablePropertyDescriptor[ epds.size() ];
	    epds.copyInto( toRet );
	    return toRet;

	}
    } // end inner class FeatureDesc


    // {{{ main

    /**
     * Testing code for the Fstruct.
     **/
    public static void main(String args[])
    {
	// Get the properties.
	Properties ourProps = new Properties();
	try {
	    ourProps = NTLLauncher.loadNTLSystemProperties( null );
	} catch (FileNotFoundException fnfe) {
	    System.out.println("Could not find file: " +
			       fnfe.getMessage() );
	    fnfe.printStackTrace();
	    System.exit(-1);
	} catch (IOException ioe) {
	    System.out.println("Could not load file: " +
			       ioe.getMessage() );
	    ioe.printStackTrace();
	    System.exit(-1);
	} // end try-catch

	String dictFN = ourProps.getProperty("ntlsys.dataDir") +
	    "/" + ourProps.getProperty("ntlsys.dictDir") +
	    "/" + ourProps.getProperty("ntlsys.featureDictName");
	FeatureDictionary fd = null;

	try {
	    ObjectInputStream ois =
		new ObjectInputStream( new FileInputStream(dictFN) );
	    fd = (FeatureDictionary)ois.readObject();
	    ois.close();
	} catch (IOException ioe) {
	    System.out.println(ioe.getMessage());
	    ioe.printStackTrace();
	} catch (ClassNotFoundException cnfe) {
	    // Won't happen... but just in case.
	    System.out.println(cnfe.getMessage());
	    cnfe.printStackTrace();
	}
    
	if (fd == null) {
	    System.out.println("Feature Dictionary _not_ restored.");
	    fd = new FeatureDictionary();
	}

	Fstruct testFS = new Fstruct("Josh's Testing Fstruct", fd);
	testFS.setProperty("Foo", "Bar");
	testFS.setProperty("Visibility", "true");
	testFS.setProperty("force.irrelevant", "");
	testFS.setProperty("test-irrelevant.irrelevant", "");
	testFS.setProperty("test-ignore.ignore", "");

	Fstruct testFS2 = new Fstruct("Josh's Second Testing Fstruct", fd);
	try {
	    Feature f = fd.getFeature("foobar");
	    testFS.putFeature( f.getName(), f );
	    f = fd.getFeature("force");
	    testFS.putFeature( f.getName(), f );
	    f = fd.getFeature("testing");
	    testFS.putFeature( f.getName(), f );
	    f = fd.getFeature("testing ifd");
	    testFS.putFeature( f.getName(), f );
	    testFS2.putFeature( f.getName(), f );
	} catch (NoSuchFeatureException nsfe) {
	    System.out.println( "Int error: " + nsfe.getMessage() );
	    //throw new InternalError( nsfe.getMessage() );
	} catch (NTLException ntle) {
	    System.out.println( "Int error: " + ntle.getMessage() );
	    //throw new InternalError( ntle.getMessage() );
	} catch (NullPointerException npe) {
	    System.out.println( "Int error: " + npe.getMessage() );
	}
	FstructViewer fsv = new FstructViewer( testFS );
	JFrame jf = new JFrame("Test");
	jf.getContentPane().add( new JScrollPane(fsv) );
	jf.pack();
	jf.setVisible( true );

	System.out.println("BEFORE EDITING...");
	System.out.println( testFS.toString() );

	WindowWaiter ww = new WindowWaiter();
	jf.addWindowListener( ww );
	ww.WaitUntilDone();

	//fsv.setTarget( testFS );

	// Edit.
	//testFS.Edit(true);

	// Wait until Editing window is done...
	System.out.println("AFTER EDITING...");
	System.out.println(testFS.toString());

	// Save the FeatureDict when done
	try {
	    ObjectOutputStream oos =
		new ObjectOutputStream( new FileOutputStream(dictFN) );
	    oos.writeObject(fd);
	    oos.flush();
	    oos.close();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.out.println(ioe.getMessage());
	}

	System.exit(0);

    }
    
    // }}} main

}

