package NTLSystem.support;

import java.io.*;
import java.util.*;

/**
 * Implement adjustable significant-digits for printing doubles.
 * <P>
 * $Id: Fmt.java,v 1.4 1998/08/08 02:02:50 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @version   $Revision: 1.4 $
 **/
public class Fmt
{

  // {{{ ATTRIBUTES
  /** Number of significant digits  **/
  private static int sigDigits = 3;

  /**
   * Set number of significant digits.
   *
   * @param     sd Number of significant digits.
   **/
  public static void setSigDigits(int sd) {
    if (sd >= 1) {
      sigDigits = sd;
    }
    else {
      System.err.println("Warning: Fmt cannot set sigDigits to " + sd);
    }
  }

  /**
   * Get number of significant digits.
   *
   * @return     Number of significant digits.
   **/
  public static int getSigDigits() { return sigDigits; }

  /**
   * Return a string version of 'd' with at most sigDigits digits
   * displayed.  Handles both exponential and conventional notation.
   * Assumes exponentials returned by double.toString() are in normal
   * form (one non-zero digit left of decimal).
   *
   * @param     d Double to be converted to String.
   * @return     string version of double with max sigDigits displayed.
   **/
  public static String dbl(double d) {
    String s = new Double(d).toString();
    if (s.equals("NaN") || s.equals("Infinity") || s.equals("-Infinity")) {
      return s;
    }
    s = s.toLowerCase();
    if (s.indexOf('e') >= 0) {
      boolean neg = (s.charAt(0) == '-');
      String negStr = (neg ? "-" : "");
      int wholeIndex = (neg ? 1 : 0);
      int expIndex = s.indexOf('e');
      String wholeStr = s.substring(wholeIndex, wholeIndex + 1);
      String fracStr = "";
      if (sigDigits > 1) {
 	fracStr = s.substring(wholeIndex + 1,
 			      Math.min(expIndex, 
 				       wholeIndex + sigDigits + 1));
      }
      String expStr = s.substring(expIndex);
      return negStr + wholeStr + fracStr + expStr;
    }
    else {
      boolean neg = (s.charAt(0) == '-');
      String negStr = (neg ? "-" : "");
      int wholeIndex = (neg ? 1 : 0);
      int decIndex = s.indexOf('.');
      if (decIndex < 0) {
	decIndex = s.length();
	s = s + ".0";
      }
      StringBuffer wholeStrB = new StringBuffer(s.substring(wholeIndex, 
							    decIndex));
      StringBuffer fracStrB = new StringBuffer(s.substring(decIndex));
      for (int i = sigDigits; i < (decIndex - wholeIndex); i++) {
 	wholeStrB.setCharAt(i, '0');
      }
      int remainDigits;
      // Look out for cases like "0.123":
      if (wholeStrB.charAt(0) == '0') {
	remainDigits = sigDigits;
      }
      else {
	remainDigits = Math.max(0, sigDigits - (decIndex - wholeIndex));
      }
      int charsToKeep = ((remainDigits > 0) ? (remainDigits + 1) : 0);
      fracStrB.setLength(Math.min(charsToKeep, s.length() - decIndex));
      return negStr + wholeStrB + fracStrB;
    }
  }

    /**
     *Read a double, even if it's in exponential notation.
     *
     * @param     s The String to read as double.
     * @return    double based on string.
     **/
  public static double getDbl(String s)
    {
	s = s.toLowerCase();
	int idx = s.indexOf('e');
	if (idx >= 0) {
	    double mantissa = Double.valueOf(s.substring(0, idx)).doubleValue();
	    int exponent = Integer.parseInt(s.substring(idx + 1));
	    return mantissa * Math.pow(10, (double)exponent);
	} else {
	    return Double.valueOf(s).doubleValue();
	}
    }
}
