package NTLSystem.support;

import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import java.io.Serializable;
import java.text.*;
import java.util.*;

/**
 * A StringSortedHashtable is a Hashtable which has String keys stored
 * in sorted order.  All lookups and dictionary operations are done
 * through the Hashtable for efficiency of lookups.  This class adds a
 * new method, <A HREF="#getSortedKeys()">getSortedKeys()</A>, which
 * returns an Enumeration of the keys in the SortedHashtable.  As
 * well, this class implements the ListModel interface, making it
 * suitable for use in a JList or other Swing component which uses a
 * ListModel.
 * <P>
 * Comparisons between keys are done using a Collator object using
 * the "en.US" locale.
 * <P>
 * $Id: StringSortedHashtable.java,v 1.5 1998/10/30 01:28:11 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @version   $Revision: 1.5 $
 **/
public class StringSortedHashtable
    extends Hashtable
    implements Serializable, Cloneable, ListModel
{
    // {{{ Attributes

    /**
     * The keys used in the Hashtable, in sorted order.
     **/
    private Vector sortedKeys;

    /**
     * The EventListenerList used for the AbstractListModel
     * implemenation.
     **/
    private transient EventListenerList listenerList;

    // }}} Attributes

    // {{{ Constructors

    /**
     * Constructs a StringSortedHashtable given the initialCapacity.
     *
     * @param     initialCapacity The initial capacity of the
     *            StringSortedHashtable.
     **/
    public StringSortedHashtable(int initialCapacity)
    {
	super( initialCapacity );
	sortedKeys = new Vector( initialCapacity );
	listenerList = new EventListenerList();
    }

    /**
     * Constructs a StringSortedHashtable given the initialCapacity,
     * and loadFactor.
     *
     * @param     initialCapacity The initial capacity of the
     *            Hashtable.
     * @param     loadFactor The load factor of the Hashtable.
     **/
    public StringSortedHashtable(int initialCapacity,
				 float loadFactor)
    {
	super(initialCapacity, loadFactor);
	sortedKeys = new Vector(initialCapacity);
	listenerList = new EventListenerList();
    }

    /**
     * Constructs an empty StringSortedHashtable given the Hashtable
     * default values.  Required constructor for serialization.
     **/
    public StringSortedHashtable()
    {
	super();
	sortedKeys = new Vector();
	listenerList = new EventListenerList();
    }

    /**
     * Constructs a StringSortedHashtable from the given initial data.
     * The keys of the initialData Hashtable will be assumed to be of
     * type String, and the keys and their data will be placed in this
     * StringSortedHashtable.
     *
     * @param     initialData A Hashtable from which to draw the
     *            initial data for this StringSortedHashtable.
     **/
    public StringSortedHashtable(Hashtable initialData)
    {
	// Construct the Hashtable with the correct size.
	super( initialData.size() );
	sortedKeys = new Vector( initialData.size() );
	listenerList = new EventListenerList();

	// Copy the elements from the initialData into this object.
	String key;
	for (Enumeration e=initialData.keys();
	     e.hasMoreElements(); ) {
	    key = (String)e.nextElement();
	    put(key, initialData.get(key));
	}
    }

    // }}} Constructors

    // {{{ Methods

    /**
     * Finds and returns in index at which the String should be inserted
     * to keep the vector in sorted order.
     *
     * @param     v The Vector [of Strings], in sorted order, to insert
     *            in. 
     * @param     s The new String to insert.
     * @return    The index to insert the new String at.
     **/
    public static int FindSortedIndex(Vector v, String s)
    {
	// Linear search through the Vector, comparing elements.
	Collator c = Collator.getInstance(new Locale("en", "US", ""));
	int retIdx=0;
	for (int i=0; i<v.size(); i++) {
	    if ( c.compare((String)v.elementAt(i), s) > 0) {
		return i;
	    }
	}
	// If we can't find a location after going through the entire
	// vector, we put it at the end. 
	return v.size();
    }

    /**
     * A redefinition of the basic Hashtable.get(Object) method.  
     *
     * @param     key The key to use for the given value.  This key
     *            should be of type String.
     * @returns   the value to which key is mapped in StringSortedHashtable.
     **/
    public synchronized Object get(Object key)
    {
      return super.get(key);
    }


    /**
     * A redefinition of the basic Hashtable.put(Object) method.  This
     * version also maintains the sorted list, and fires the requisite
     * changes on any registered ListDataEvent listener.
     *
     * @param     key The key to use for the given value.  This key
     *            should be of type String.
     * @param     value The value to store in the Hashtable.
     **/
    public synchronized Object put(Object key, Object value)
    {
	Object toRet = super.put(key, value);
	if (sortedKeys != null) {
	    int idx = StringSortedHashtable.FindSortedIndex( sortedKeys,
							     (String)key );
	    sortedKeys.insertElementAt(key, idx);
	    // Update the ListModel
	    fireIntervalAdded(this, idx, idx);
	}
	return toRet;
    }

    /**
     * Removes the Object in the hashtable with a key of key, given as
     * a parameter.  The key is also removed, of course, from the
     * sortedKeys list.  The appropriate ListModel update is made.
     *
     * @param     key The key of the Object to remove from the
     *            StringSortedHashtable. 
     * @return    The removed value.
     **/
    public synchronized Object remove(Object key)
    {
	Object toRet = super.remove(key);
	// Find it in the sortedKeys Vector
	int idx = sortedKeys.indexOf( key );
	// Remove the key in the sortedKeys Vector
	sortedKeys.removeElementAt( idx );
	// Update the ListModel
	fireIntervalRemoved( this, idx, idx );
	return toRet;
    }

    /**
     * Gets the keys used in the Hashtable in String-sorted order.
     *
     * @return    A sorted Enumeration of the keys used in the
     *            Hashtable. 
     **/
    public Enumeration getSortedKeys()
    {
	return sortedKeys.elements();
    }

    // {{{ Cloneable implemenation
    
    public Object clone()
    {
	StringSortedHashtable toRet = (StringSortedHashtable)super.clone();
	toRet.sortedKeys = (Vector)sortedKeys.clone();
	// The event listener list is not cloneable, so we must do it
	// by hand. 
	toRet.listenerList = new EventListenerList();
	Object[] listeners = listenerList.getListenerList();
	for (int i=0; i<listeners.length; i+=2) {
	    toRet.listenerList.add( (Class)listeners[i],
				    (EventListener)listeners[i+1] );
	} // end for over i
	return toRet;
    }

    // }}} Cloneable implemenation


    // {{{ ListModel implementation

    /*
     * The ListModel implementation is modeled after the
     * AbstractListModel interface... however, we can't subclass it
     * since we're already subclassing Hashtable... but it's pretty
     * simple, so we just re-implement it...
     */

    public void addListDataListener(ListDataListener l)
    {
	if (listenerList == null) {
	    listenerList = new EventListenerList();
	}
	listenerList.add(ListDataListener.class, l);
    }

    protected void fireContentsChanged(Object source,
				       int index0,
				       int index1)
    {
	ListDataEvent ccEvent = new ListDataEvent(source,
						  ListDataEvent.CONTENTS_CHANGED,
						  index0, index1);
	if (listenerList == null) {
	    listenerList = new EventListenerList();
	}
	Object[] listeners = listenerList.getListenerList();
	for (int i=listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == ListDataListener.class) {
		((ListDataListener)listeners[i+1]).contentsChanged( ccEvent );
	    } // end if
	} // end for over i
    }

    protected void fireIntervalAdded(Object source,
				     int index0,
				     int index1)
    {
	ListDataEvent iaEvent = new ListDataEvent(source,
						  ListDataEvent.INTERVAL_ADDED,
						  index0, index1);
	if (listenerList == null) {
	    listenerList = new EventListenerList();
	}
	Object[] listeners = listenerList.getListenerList();
	for (int i=listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == ListDataListener.class) {
		((ListDataListener)listeners[i+1]).intervalAdded( iaEvent );
	    } // end if
	} // end for over i
    }

    protected void fireIntervalRemoved(Object source,
				       int index0,
				       int index1)
    {
	ListDataEvent irEvent = new ListDataEvent(source,
						  ListDataEvent.INTERVAL_REMOVED,
						  index0, index1);
	if (listenerList == null) {
	    listenerList = new EventListenerList();
	}
	Object[] listeners = listenerList.getListenerList();
	for (int i=listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == ListDataListener.class) {
		((ListDataListener)listeners[i+1]).intervalRemoved( irEvent );
	    } // end if
	} // end for over i
    }
    
    public Object getElementAt(int index)
    {
	return sortedKeys.elementAt(index);
    }

    public int getSize()
    {
	return sortedKeys.size();
    } 

    public void removeListDataListener(ListDataListener l)
    {
	if (listenerList == null) {
	    listenerList = new EventListenerList();
	}
	listenerList.remove(ListDataListener.class, l);
    }

    // }}} ListModel implementation

    // }}} Methods

}

