package NTLSystem.verblearn;

import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyVetoException;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * The LexiconViewer allows for the vieweing of a Lexicon, including
 * the Words, their Senses and the Fstructs resultant from training.
 * A LexiconViewer is associated with a single Lexicon.  If the
 * Lexicon is modiifed while the viewer is open, changes won't
 * necessarily be reflected until a new LexiconViewer is opened, or
 * the LexiconViewer is retargeted [with 
 * <A HREF="#setTarget(Lexicon)">setTarget()</A>] to the updated
 * Lexicon.
 * <P>
 * $Id: LexiconViewer.java,v 1.5 1998/10/16 01:59:35 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 * 
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Lexicon
 * @version   $Revision: 1.5 $
 **/
public class LexiconViewer
    extends JInternalFrame
    implements ListSelectionListener
{
    // {{{ Attributes

    /** The VerbLearn instance which contains this LexiconViewer. **/
    private VerbLearn vl;

    /** The Lexicon being viewed. **/
    private Lexicon l;

    /** The JList of Words in the targeted Lexicon. **/
    private JList wordList;
    /** The JScrollPane which contains the wordList. **/
    private JScrollPane wordListScroller;

    /** The JList of Senses in the currently-selected Word. **/
    private JList senseList;
    /** The JScrollPane which contains the senseList. **/
    private JScrollPane senseListScroller;

    /** A JList of the virtual sample values for this Lexicon. **/
    private JList vsampList;
    /** The JScrollPane which contains the vsampList. **/
    private JScrollPane vsampListScroller;

    /** The number of rows to display in the word and sense
     * lists. **/
    private static final int LIST_ROW_COUNT = 10;

    /** The width of the word and sense lists. **/
    private static final int LIST_WIDTH = 100;

    /** The JPanel which contains the FstructViewers viewing the
	Fstructs which make up the particular selected Sense. **/
    private JPanel fstructs;

    /** The JLabel containing detail of how many examples are covered
	by the current Sense. **/
    private JLabel exampleCount;

    /** The JTextField what Language the Lexicon is in. **/ 
    private JLabel lexLang;

    /** The JLabel which which displays and allows editing of what the
	name of the Lexicon is. **/ 
    private JTextField lexName;

    /** The JLabel which indicates how many Words and Senses are in
	the Lexicon.  **/
    private JLabel lexContains;

    /** The title stem of this Window. **/
    private static String TITLE_STEM = "Lexicon Viewer - ";

    // }}} Attributes

    // {{{ Constructors

    /**
     * Constructs a LexiconViewer given a VerbLearn instance which
     * contains it and a Lexicon to view.
     *
     * @param     vl The VerbLEarn instance which contains this
     *            LexiconViewer.
     * @param     l The Lexicon being viewed.
     **/
    public LexiconViewer(VerbLearn vl, Lexicon l)
    {
	super( TITLE_STEM + l.getName(),
	       true, true, true, true    );

	this.vl = vl;
	this.l = l;

	JComponent jc = (JComponent)getContentPane();
	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();

	jc.setLayout( gbl );

	// Give some general detail about the lexicon 
	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridwidth = GridBagConstraints.REMAINDER;
	JPanel jp = new JPanel( new GridBagLayout() );

	lexName = new JTextField( " (empty name) ", 20 );
	lexName.addActionListener( new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		// Check to see if the name is already used.
		if ( LexiconViewer.this.vl.getScenarioSet().
		     getLexicon( lexName.getText() ) != null ) {
		    JOptionPane.
			showMessageDialog( LexiconViewer.this.vl,
					   "The name " +
					   lexName.getText() +
					   " is already in use.",
					   "Name already in use.",
					   JOptionPane.ERROR_MESSAGE );
		} else {
		    String newName = lexName.getText();
		    //  change the name
		    LexiconViewer.this.l.setName( newName );
		    //  change the titlebar
		    setTitle( TITLE_STEM + newName );
		    //  rebuild the lexicon menu
		    LexiconViewer.this.vl.buildLexiconMenu();
		} // end if-else
	    }
	} );
	lexName.setBorder( BorderFactory.
			   createTitledBorder( "Name" ) );
	((GridBagLayout)jp.getLayout()).setConstraints( lexName, c );
	jp.add( lexName );
	
	lexLang = new JLabel( " (empty language) ",
			      JLabel.CENTER );
	lexLang.setBorder( BorderFactory.
			   createTitledBorder( "Language" ) );
	((GridBagLayout)jp.getLayout()).setConstraints( lexLang, c );
	jp.add( lexLang );
	
	// Use a dummy string which should create a label of
	// sufficient length.
	lexContains = new JLabel( "99 Words covering 999 Senses.",
				  JLabel.CENTER );
	lexContains.
	    setBorder( BorderFactory.
		       createCompoundBorder(BorderFactory.
					    createTitledBorder( "Contains" ),
					    BorderFactory.
					    createEmptyBorder(3, 3, 3, 3)) );
	((GridBagLayout)jp.getLayout()).setConstraints( lexContains,
							c );
	jp.add ( lexContains );

	c.fill = GridBagConstraints.HORIZONTAL;
	JButton jb = new JButton( "Remove Lexicon" );
	jb.addActionListener( new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		// Remove this Lexicon...
		// Confirm with user.
		if ( JOptionPane.
		     showConfirmDialog(LexiconViewer.this.vl,
				       "Remove Lexicon " +
				       LexiconViewer.this.l.getName() + "?",
				       "Remove this Lexicon?",
				       JOptionPane.YES_NO_OPTION)
		     == JOptionPane.YES_OPTION) {
		    // If so..
		    //  remove Lexicon from ScenarioSet.
		    LexiconViewer.this.vl.getScenarioSet().
			removeLexicon( LexiconViewer.this.l.getName() );
		    //  Rebuild menu.
		    LexiconViewer.this.vl.buildLexiconMenu();

		    //  If there are now no lexicons, remove the
		    //  Testing menu items.
		    if ( LexiconViewer.this.vl.getScenarioSet().
			 getLexicons().size() == 0 ) {
			LexiconViewer.this.vl.obey.
			    setEnabled( false );
			LexiconViewer.this.vl.recognize.
			    setEnabled( false );
		    } // end if

		    // And close ourself.
		    try {
			LexiconViewer.this.setClosed( true );
		    } catch (PropertyVetoException pve) {
			// Shouldn't happen.
			throw new InternalError( pve.getMessage() );
		    } // end try-catch
		} 
	    }
	} );
	((GridBagLayout)jp.getLayout()).
	    setConstraints( jb, c );
	jp.add( jb );

	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = 1;
	c.anchor = GridBagConstraints.NORTH;
	gbl.setConstraints( jp, c );
	jc.add( jp );

	// Create a Panel for the Word list.
	jp = new JPanel( new GridBagLayout() );
	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridwidth = GridBagConstraints.REMAINDER;
	JLabel jl = new JLabel( "Words", JLabel.CENTER );
	((GridBagLayout)jp.getLayout()).setConstraints( jl, c );
	jp.add( jl );

	c.fill = GridBagConstraints.BOTH;
	wordList = new JList();
	wordList.setVisibleRowCount( LIST_ROW_COUNT );
	wordList.setFixedCellWidth( LIST_WIDTH );
	wordList.setSelectionMode( ListSelectionModel.
				   SINGLE_SELECTION );
	wordList.addListSelectionListener( this );
	wordListScroller =
	    new JScrollPane( wordList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
	((GridBagLayout)jp.getLayout()).
	    setConstraints( wordListScroller, c );
	jp.add( wordListScroller );

	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = 1;
	gbl.setConstraints( jp, c );
	jc.add( jp );

	// Create a panel for the Sense list.
	jp = new JPanel( new GridBagLayout() );
	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridwidth = GridBagConstraints.REMAINDER;

	jl = new JLabel( "Senses", JLabel.CENTER );
	((GridBagLayout)jp.getLayout()).setConstraints( jl, c );
	jp.add( jl );

	c.fill = GridBagConstraints.BOTH;
	senseList = new JList();
	senseList.setVisibleRowCount( LIST_ROW_COUNT );
	senseList.setFixedCellWidth( LIST_WIDTH );
	senseList.setSelectionMode( ListSelectionModel.
				    SINGLE_SELECTION );
	senseList.addListSelectionListener( this );
	senseListScroller =
	    new JScrollPane( senseList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
	((GridBagLayout)jp.getLayout()).setConstraints( senseList,
							c );
	jp.add( senseListScroller );
	
	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = GridBagConstraints.RELATIVE;
	gbl.setConstraints( jp, c );
	jc.add( jp );

	// Add the virtual sample list
	jp = new JPanel( new GridBagLayout() );
	
	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridwidth = GridBagConstraints.REMAINDER;
	jl = new JLabel( "Virtual Samples", JLabel.CENTER );
	((GridBagLayout)jp.getLayout()).setConstraints( jl, c );
	jp.add( jl );
	
	c.fill = GridBagConstraints.BOTH;
	vsampList = new JList();
	vsampList.setVisibleRowCount( LIST_ROW_COUNT );
	vsampList.setFixedCellWidth( 2 * LIST_WIDTH );
	vsampListScroller =
	    new JScrollPane( vsampList,
			     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
			     JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );

	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = GridBagConstraints.REMAINDER;
	((GridBagLayout)jp.getLayout()).
	    setConstraints( vsampListScroller, c );
	jp.add( vsampListScroller );

	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints( jp, c );
	jc.add( jp );

	// Add the Fstruct viewer
	c.fill = GridBagConstraints.BOTH;

	jp = new JPanel( new GridBagLayout() );
	jp.setBorder( BorderFactory.
		      createTitledBorder("Sense detail") );
	
	c.fill = GridBagConstraints.HORIZONTAL;
	c.gridwidth = GridBagConstraints.REMAINDER;

	// Add something that shows how many examples cover the
	// currently-selected sense.
	String senseName = (String)senseList.getSelectedValue();
	if (senseName == null) {
	    senseName = " (unselected) ";
	}
	exampleCount = new JLabel( "Sense:           ( 00 Examples )",
				   JLabel.CENTER );
	((GridBagLayout)jp.getLayout()).setConstraints( exampleCount,
							c );
	jp.add( exampleCount );
	
	fstructs = new JPanel();
	Vector templates = l.getScenarioSet().getTemplateFstructs();
	FstructViewer fsvToAdd;
	for (int i=0; i<templates.size(); i++) {
	    fsvToAdd = new FstructViewer( null );
	    fsvToAdd.setBorder( BorderFactory.
				createLineBorder( Color.black ) );
	    fstructs.add( fsvToAdd );
	} // end for over template fstructs
	fstructs.setSize( 700, 450 );

	c.fill = GridBagConstraints.BOTH;
	JScrollPane jsp =
	    new JScrollPane( fstructs,
			     JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
			     JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS );
	((GridBagLayout)jp.getLayout()).setConstraints( jsp, c );
	jp.add( jsp );

	c.fill = GridBagConstraints.BOTH;
	c.gridwidth = GridBagConstraints.REMAINDER;
	c.weightx = c.weighty = 1.0;
	gbl.setConstraints( jp, c );
	jc.add( jp );
	
	// Done... pack what we have.
	pack();

	// Fill in the fields.
	setTarget( l );

	// Make the window visible.
	setVisible( true );

    }

    // }}} Constructors

    // {{{ Methods

    // {{{ setTarget

    public void setTarget(Lexicon l)
    {
	// Fix the window title.
	setTitle( TITLE_STEM + l.getName() );
	// Fix the language label.
	lexLang.setText( l.getLanguage() );
	lexLang.validate();
	// Fix the name label.
	lexName.setText( l.getName() );
	lexName.validate();
	// Fix the contains label.
	updateWordSenseCount();
	// Fix the Word list.
	wordList.clearSelection();
	// Get the new Word list [these are returned in sorted order.]
	Enumeration we = l.getWords();
	Vector wv = new Vector( l.wordCount() );
	String toInsert;
	while ( we.hasMoreElements() ) {
	    toInsert = (String)((Word)we.nextElement()).getName();
	    wv.addElement( toInsert );
	} // end while
	wordList.clearSelection();
	wordList.setListData( wv );
	wordListScroller.validate();
	wordListScroller.update( wordListScroller.getGraphics() );
	// Fix the Sense list.
	senseList.clearSelection();
	senseList.setListData( new Vector() );
	senseListScroller.validate();
	senseListScroller.update( senseListScroller.getGraphics() );
	// Fix the virtual sample list
	vsampList.clearSelection();
	updateVirtSampList();
	// Fix the Fstruct viewer.
	Component[] cs = fstructs.getComponents();
	for (int i=0; i<cs.length; i++) {
	    ((FstructViewer)cs[i]).setTarget( null );
	} // end for over fstructs components [FstructViewers]
    }
    // }}} setTarget

    // {{{ updateWordSenseCount

    /**
     * Updates the label which contains the number of Words and Senses
     * which the Lexicon contains.
     **/
    protected void updateWordSenseCount()
    {
	int wc = l.wordCount();
	int sc = l.senseCount();
	String wordDesc = (wc == 1) ? " Word" : " Words";
	String senseDesc = (sc == 1) ? " Sense" : " Senses";
	lexContains.setText( wc + wordDesc +
			     " covering " +
			     sc + senseDesc + "." );
	lexContains.validate();
    }

    // }}} updateWordSenseCount

    // {{{ updateVirtSampList

    protected void updateVirtSampList()
    {
	Enumeration vsEnum = l.getVirtualSampleFeatures();
	Vector vsVect = new Vector( );
	String vsName;
	double vsVal;
	int sigDigs;
	sigDigs = Fmt.getSigDigits();
	Fmt.setSigDigits( 5 );
	while ( vsEnum.hasMoreElements() ) {
	    vsName = (String)vsEnum.nextElement();
	    vsVal = l.getVirtualSamples( vsName );
	    
	    vsVect.addElement( vsName + "=" +
			       Fmt.dbl( vsVal ) );
	} // end while over virtual samples
	vsampList.setListData( vsVect );
	vsampList.validate();
	// Force the visible portion of the list to redraw since
	// validate() won't do it...
	vsampList.update( vsampList.getGraphics() );
	Fmt.setSigDigits( sigDigs );
    }
    
    // }}} updateVirtSampList

    // {{{ ListSelectionListener impl.

    public void valueChanged(ListSelectionEvent lse)
    {
	if ( lse.getValueIsAdjusting() )
	    return;

	JList list = (JList)lse.getSource();
	if (list == wordList) {
	    handleWordListSelection();
	} else if (list == senseList) {
	    handleSenseListSelection();
	}
    }

    public void handleWordListSelection()
    {
	// Get the selected Word.
	String selWordName =
	    (String)wordList.getSelectedValue();
	Word selWord = l.getWord( selWordName );
	if (selWord != null) {
	    // Update the Sense list
	    senseList.clearSelection();
	    Enumeration se = selWord.getSenses();
	    Vector sv = new Vector( selWord.senseCount() );
	    Sense s;
	    int insIdx;
	    while ( se.hasMoreElements() ) {
		s = (Sense)se.nextElement();
		// Insert the Sense names in sorted order for coolness.
		insIdx = StringSortedHashtable.FindSortedIndex( sv,
								s.getName() );
		sv.insertElementAt( s.getName(), insIdx );
	    } // end while over se
	    senseList.setListData( sv );
	    senseListScroller.validate();
	    senseListScroller.update( senseListScroller.getGraphics() );
	}
    }

    public void handleSenseListSelection()
    {
	String selWordName = (String)wordList.getSelectedValue();
	if (selWordName == null)
	    return;
	Word selWord = l.getWord( selWordName );
	if (selWord == null)
	    return;
	String selSenseName = (String)senseList.getSelectedValue();
	if (selSenseName == null)
	    return;
	// Select the correct Sense from the Word by going through the
	// Word's Enumeration of Senses.
	Enumeration e = selWord.getSenses();
	boolean notFound = true;
	Sense s = null;
	while ( e.hasMoreElements() && notFound ) {
	    s = (Sense)e.nextElement();
	    if ( s.getName().equals( selSenseName ) ) {
		notFound = false;
	    }
	} // end while over senses [enumeration]
	if (notFound) {
	    throw new InternalError("Can't find Sense " +
				    selSenseName +
				    " expected to be in Word " +
				    selWordName +
				    " in Lexicon " + l.getName() );
	}

	// Update the example count
	exampleCount.setText("Sense: " + s.getName() +
			     " (" + s.numExamples() +
			     " examples)");
	exampleCount.validate();

	// Update the FstructViewers
	Component[] cs = fstructs.getComponents();
	Fstruct[] fs = s.splitFstructs( vl.getFeatureDict() );
	// assertion...
	if (cs.length != fs.length) {
	    throw new InternalError("FstructViewer[] length: " +
				    cs.length +
				    ", Fstruct[] length: " +
				    fs.length );
	}
	// Retarget the FstructViewers
	System.out.println("Retargeting " + cs.length +
			   " FstructViewers");
	for (int i=0; i<cs.length; i++) {
	    System.out.println("  Retargeting FV " + i );
	    ((FstructViewer)cs[i]).setTarget(fs[i]);
	    fstructs.validate();
	}
    }

    // }}} ListSelectionListener impl.

    // }}} Methods


}
