package NTLSystem.verblearn;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Properties;
import java.util.Vector;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * The main VerbLearn program.  If started as the main object, it will
 * load a blank, new VerbLearn program instance.  In addition,
 * a VerbLearn instance can be created directly from another program
 * with the constructors.
 * <P>
 *
 * $Id: VerbLearn.java,v 1.10 1998/10/16 01:59:43 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       ScenarioSetViewer
 * @see       ScenarioViewer
 * @version   $Revision: 1.10 $
 **/
public class VerbLearn
    extends JFrame
{

    // {{{ Attributes

    /**
     * The VerbLearn properties.  This should be set before a
     * VerbLearn instance is created.
     **/
    protected static Properties vlProps;
    /**
     * The VerbLearn FeatureDictionary.  This should be set before a
     * VerbLearn instance is created, otherwise the blank [new]
     * FeatureDictionary will be used.
     **/
    protected static FeatureDictionary featureDict = new FeatureDictionary();

    private static final int WIDTH = 1000;
    private static final int HEIGHT = 768;

    protected JMenuBar menubar;
    protected JMenu fileMenu;
    protected JMenu scenarioSetMenu;
    protected JMenu lexiconMenu;
    protected JMenu openLexiconMenu;
    protected JMenu testMenu;
    
    protected JDesktopPane desktop;

    /**
     * The currently-loaded ScenarioSet.  If null, this instance of
     * VerbLearn is not currently manipulating a ScenarioSet.
     **/
    private ScenarioSet sset;

    /**
     * The filename of the currently-loaded ScenarioSet.  If null, the
     * currently-loaded ScenarioSet doesn't have a file name, which
     * should be taken into account for "Save Scenario Set"...
     **/
    private File ssetFile;

    // {{{ Actions

    /**
     * Creates a new ScenarioSet.
     **/
    private Action newSSet;
    /**
     * Loads a ScenarioSet.
     **/
    private Action loadSSet;
    /**
     * Saves the currently-loaded ScenarioSet.  If a filename doesn't
     * exist for the loaded ScenarioSet, saveSSetAs will be invoked.
     **/
    private Action saveSSet;
    /**
     * Saves the currently-loaded ScenarioSet with the specific
     * filename. 
     **/
    private Action saveSSetAs;
    /**
     * Exits the program after checking if the loaded ScenarioSet needs
     * to be saved.
     **/
    private Action exit;
    /**
     * Pops up an 'about' dialog box.
     **/
    private Action about;

    private Action openSSetViewer;
    private Action openViewer;
    private Action generate;
    private Action train;
    private Action lexiconAnalysis;
    protected Action recognize;
    protected Action obey;

    /** The window title stem.  The name of the opened ScenarioSet is
	appended to this. **/
    private static final String WINDOW_TITLE_STEM = "VerbLearn - ";

    // }}} Actions
    
    // }}} Attributes
    
    // {{{ Constructors

    /**
     * Creates the basic VerbLearn GUI instance.  Creates the JFrame,
     * the MenuBar and associated actions, etc.
     **/
    public VerbLearn()
    {
	super(WINDOW_TITLE_STEM + "<untitled>");
	// Initialize variables.
	sset = null;
	ssetFile = null;
	
	desktop = new JDesktopPane();
	getContentPane().add(desktop);
	
	// Create the menus.
	createMenus();
	// Initialize the Actions
	saveSSet.setEnabled( false );
	saveSSetAs.setEnabled( false );
	openSSetViewer.setEnabled( false );
	openViewer.setEnabled( false );
	generate.setEnabled( false );
	train.setEnabled( false );
	recognize.setEnabled( false );
	obey.setEnabled( false );
	
	// Add the menu bar.
	setJMenuBar(menubar);
	
	// Pack the window, resize it, and ensure it's visibility.
	pack();
	setSize( new Dimension( WIDTH, HEIGHT ) );
	setVisible(true);
    }

    /**
     * Creates a new VerbLearn instance by loading the ScenarioSet
     * from the given filename.  If the file cannot be loaded, the
     * VerbLearn instance will revert to a new ScenarioSet after 
     * informing the user of the error.
     *
     * @param     fileName The filename of the ScenarioSet to load.
     **/
    public VerbLearn(String fileName)
    {
	this();
	// Load the ScenarioSet from the filename.
	loadScenarioSet( new File(fileName) );
    }

    /**
     * Creates a new VerbLearn instance using the given ScenarioSet.
     *
     * @param     sset The ScenarioSet to manipulate.
     **/
    public VerbLearn(ScenarioSet sset)
    {
	this();
	this.sset = sset;
	// Configure available actions.
	saveSSet.setEnabled( true );
	saveSSetAs.setEnabled( true );
	openSSetViewer.setEnabled( true );
	openViewer.setEnabled( true );
	generate.setEnabled( true );
	if (sset.countScenarios() > 0) {
	    train.setEnabled( true );
	} else {
	    train.setEnabled( false );
	}
	if ( sset.getLexicons().size() > 0 ) {
	    recognize.setEnabled( true );
	    obey.setEnabled( true );
	}  else {
	    recognize.setEnabled( false );
	    obey.setEnabled( false );
	}
	// Set the window title
	setTitle( WINDOW_TITLE_STEM + sset.getName() );
    }

    // }}} Constructors

    // {{{ Methods
    
    // {{{ createMenus()
    /**
     * Creates the menus for the VerbLearn application.  In reality,
     * it does this by creating a bunch of AbstractAction objects,
     * which are then placed in the menu bar.  These Action objects
     * can be used, as well, for a toolbar... but they're not.
     * However, the Actions have the nice ability that they can be
     * disabled, and their menu item and toolbar item [if someone
     * wants to add a toolbar later] will both be disabled.
     **/
    protected void createMenus()
    {
	menubar = new JMenuBar();
	JMenu jm;
	JMenuItem jmi;

	// Create the "File" menu.
	fileMenu = new JMenu("File");
	
	newSSet = new AbstractAction("New Scenario Set") {
	    public void actionPerformed(ActionEvent ae) {
		// If a SSet is currently loaded, check if it should
		// be saved.
		if ( sset != null ) {
		    // FIXME: Check with user.
		} 
		// Reset variables and options.
		sset = new ScenarioSet();
		ssetFile = null;
		openSSetViewer.setEnabled( true );
		openViewer.setEnabled( true );
		generate.setEnabled( true );
		saveSSet.setEnabled( true );
		saveSSetAs.setEnabled( true );
		if ( sset.countScenarios() > 0 ){
		    train.setEnabled( true );
		} else {
		    // Disable unuseable actions.
		    train.setEnabled( false );
		}
		recognize.setEnabled( false );
		obey.setEnabled( false );
		// User should have a blank, new ScenarioSet to deal with.
	    }
	};
	fileMenu.add( newSSet );
	
	loadSSet = new AbstractAction("Load Scenario Set...") {
	    public void actionPerformed(ActionEvent ae) {
		// Open a File-selection dialog to open a scenario-set file.
		String dir =
		vlProps.getProperty("ntlsys.ntlhome") + "/" +
		vlProps.getProperty("ntlsys.verblearn.homeDir") +
		vlProps.getProperty("ntlsys.verblearn.dataDir") +
		vlProps.getProperty("ntlsys.verblearn.ssetDir");
		JFileChooser jfc = new JFileChooser( dir );
		int retval = jfc.showOpenDialog( VerbLearn.this );
		if (retval == JFileChooser.APPROVE_OPTION) {
		    final File toOpen = jfc.getSelectedFile();
		    if (toOpen != null) {
			// Read in selected file, and do it on a
			// seperate thread.
			Thread t = new Thread() {
			    public void run() {
				loadScenarioSet( toOpen );
				JOptionPane.showMessageDialog(VerbLearn.this,
							      "Scenario Set has been loaded.",
							      "Load Complete",
							      JOptionPane.INFORMATION_MESSAGE);
			    } // end Thread.run()
			}; // end Thread
			t.start();
		    }
		} else {
		    // No file selected
		    JOptionPane.showMessageDialog(VerbLearn.this,
						  "No selected file",
						  "No file has been selected.",
						  JOptionPane.INFORMATION_MESSAGE);
		}
	    }
	};
	fileMenu.add(loadSSet);
	
	saveSSetAs = new AbstractAction("Save Scenario Set As...") {
	    public void actionPerformed(ActionEvent ae) {
		// Allow the user to select a file name with which to save.
		String dir =
		vlProps.getProperty("ntlsys.ntlhome") + "/" +
		vlProps.getProperty("ntlsys.verblearn.homeDir") +
		vlProps.getProperty("ntlsys.verblearn.dataDir") +
		vlProps.getProperty("ntlsys.verblearn.ssetDir");
		JFileChooser jfc = new JFileChooser( dir );
		int retval = jfc.showSaveDialog( VerbLearn.this );
		if (retval == JFileChooser.APPROVE_OPTION) {
		    final File toSave = jfc.getSelectedFile();
		    if (toSave != null) {
			// Save the Scenario set to the given
			// filename, and do it on a seperate thread...
			// and tell the user when done.
			Thread t = new Thread() {
			    public void run() {
				saveScenarioSet( toSave );
				JOptionPane.showMessageDialog(VerbLearn.this,
							      "Scenario Set has been saved.",
							      "Save Complete",
							      JOptionPane.INFORMATION_MESSAGE );
			    } // end Thread.run
			}; // end Thead
			t.start();
		    }
		}
	    }
	};
	saveSSetAs.setEnabled(false);
	
	saveSSet = new AbstractAction("Save Scenario Set...") {
	    public void actionPerformed(ActionEvent ae) {
		// If it doesn't exist, allow user to select filename with
		// which to save.
		if (ssetFile == null) {
		    saveSSetAs.actionPerformed(ae);
		} else {
		    // Save the ScenarioSet to the already-set
		    // filename.
		    saveScenarioSet( ssetFile );
		}
	    }
	};
	saveSSet.setEnabled(false);

	fileMenu.add( saveSSet );
	fileMenu.add( saveSSetAs );
    
	fileMenu.add( new JSeparator() );

	about = new AbstractAction("About...") {
	    public void actionPerformed(ActionEvent ae) {
		// Pop up an 'about' dialog.
		new AboutDialog(VerbLearn.this,
				"VerbLearn",
				"1.0",
				"Authors: nchang, jsled, oliver, dbailey");
	    }
	};
	fileMenu.add( about );

	exit = new AbstractAction("Exit") {
	    public void actionPerformed(ActionEvent ae) {
		// FIXME: If file is loaded ... check if the user
		// wants to save it.
		dispatchEvent(new WindowEvent(VerbLearn.this,
					      WindowEvent.WINDOW_CLOSING));
	    }
	};
	fileMenu.add( exit );

	menubar.add( fileMenu );

	// Scenario Set menu
	scenarioSetMenu = new JMenu("Scenario Set");

	openSSetViewer = new AbstractAction("Open Scenario Set Options") {
	    public void actionPerformed(ActionEvent ae) {
		Thread t = new Thread() {
		    public void run() {
			ScenarioSetViewer ssv = new ScenarioSetViewer( sset );
			desktop.add(ssv, JLayeredPane.DEFAULT_LAYER);
			desktop.moveToFront( ssv );
		    } // end Thread.run
		}; // end Thread
		t.start();
	    }
	};
	scenarioSetMenu.add( openSSetViewer );

	openViewer = new AbstractAction("Open Scenario Viewer") {
	    public void actionPerformed(ActionEvent ae) {
		Thread t = new Thread() {
		    public void run() {
			ScenarioViewer sv = new ScenarioViewer( sset );
			desktop.add(sv, JLayeredPane.DEFAULT_LAYER);
			desktop.moveToFront( sv );
		    } // end Thread.run
		}; // end Thread
		t.start();
	    }
	};
	scenarioSetMenu.add( openViewer );

	generate = new AbstractAction("Generate Scenarios") {
	    public void actionPerformed(ActionEvent ae) {
		GenerateGUI gGUI = new GenerateGUI();
		desktop.add(gGUI, JLayeredPane.DEFAULT_LAYER);
		desktop.moveToFront( gGUI );
	    }
	};
	scenarioSetMenu.add(generate);

	menubar.add(scenarioSetMenu);

	lexiconMenu = new JMenu("Lexicon");

	train = new AbstractAction("Train") {
	    public void actionPerformed(ActionEvent ae) {
		TrainingGUI tGUI = new TrainingGUI(sset, VerbLearn.this);
		desktop.add(tGUI, JLayeredPane.DEFAULT_LAYER);
		desktop.moveToFront( tGUI );
	    }
	};
	lexiconMenu.add( train );
	
	lexiconMenu.add( new JSeparator() );

	lexiconAnalysis = new AbstractAction("Anaylsis...") {
	    public void actionPerformed(ActionEvent ae) {
		// FIXME: Perform Analysis
	    }
	};
	lexiconAnalysis.setEnabled( false );
	lexiconMenu.add(lexiconAnalysis);

	lexiconMenu.add( new JSeparator() );

	openLexiconMenu = new JMenu("Open Lexicon");

	// Build lexicon menu from list of available lexicons.
	buildLexiconMenu( );

	lexiconMenu.add(openLexiconMenu);

	menubar.add( lexiconMenu );

	testMenu = new JMenu( "Testing" );

	recognize = new AbstractAction( "Recognize" ) {
	    public void actionPerformed(ActionEvent ae) {
		RecognizeGUI rgui = new RecognizeGUI(VerbLearn.this,
						     VerbLearn.this.sset);
		desktop.add( rgui,
			     JLayeredPane.DEFAULT_LAYER );
		desktop.moveToFront( rgui );
	    }
	};
	recognize.setEnabled( false );
	testMenu.add( recognize );

	obey = new AbstractAction( "Obey" ) {
	    public void actionPerformed(ActionEvent ae) {
		ObeyGUI ogui = new ObeyGUI(VerbLearn.this,
						     VerbLearn.this.sset);
		desktop.add( ogui,
			     JLayeredPane.DEFAULT_LAYER );
		desktop.moveToFront( ogui );
	    }
	};
	obey.setEnabled( false );
	testMenu.add( obey );

	menubar.add( testMenu );
    
    }
    // }}} createMenus()

    // {{{ loadScenarioSet()

    /**
     * Loads a ScenarioSet from the given File and sets up this
     * instance of VerbLearn for editing that ScenarioSet.  If the
     * File cannot be loaded, the user will be notified and VerbLearn
     * will remain unchanged.
     *
     * @param     file The File corresponding to the ScenarioSet to
     *            load. 
     **/
    protected void loadScenarioSet(File file)
    {
	// Simple check.
	if (file == null) {
	    JOptionPane.showMessageDialog(this, 
					  "No specified filename.",
					  "No Filename",
					  JOptionPane.ERROR_MESSAGE);
	    return;
	}
	// Create a Stream to the file.
	try {
	    ScenarioSet tmpSSet;
	    ObjectInputStream ois =
		new ObjectInputStream(new BufferedInputStream( new FileInputStream( file ) ) );
 
	    // We do the following in two steps so if an IOException
	    // or ClassNotFoundException occurs, the already-loaded
	    // data won't be destroyed.
	    tmpSSet = (ScenarioSet)ois.readObject();
	    sset = tmpSSet;
	    ssetFile = file;
	    ois.close();
	} catch (FileNotFoundException fnfe) {
	    JOptionPane.showMessageDialog(this,
					  "Cannot find file " +
					  file.getAbsolutePath(),
					  "Cannot find file.",
					  JOptionPane.ERROR_MESSAGE);
	} catch (IOException ioe) {
	    Object msgs[] = new String[2];
	    msgs[0] = new String("Cannot load file " +
				 file.getAbsolutePath() );
	    msgs[1] = new String("IOException message: " + ioe.getMessage());
	    JOptionPane.showMessageDialog(this,
					  msgs,
					  "Cannot load file.",
					  JOptionPane.ERROR_MESSAGE);
	} catch (ClassNotFoundException cnfe) {
	    Object msgs[] = new String[2];
	    msgs[0] = new String("Cannot load file " +
				 file.getAbsolutePath() );
	    msgs[1] = new String("ClassNotFoundException message: " + cnfe.getMessage());
	    JOptionPane.showMessageDialog(this,
					  msgs,
					  "Cannot load file.",
					  JOptionPane.ERROR_MESSAGE);
	} // end try-catch-catch

	// Now that the ScenarioSet is loaded, configure the Actions
	// which are enabled|disabled.
	saveSSet.setEnabled( true );
	saveSSetAs.setEnabled( true );
	openSSetViewer.setEnabled( true );
	openViewer.setEnabled( true );
	generate.setEnabled( true );
	if (sset.countScenarios() > 0) {
	    train.setEnabled( true );
	} else {
	    train.setEnabled( false );
	}
	if ( sset.getLexicons().size() > 0 ) {
	    recognize.setEnabled( true );
	    obey.setEnabled( true );
	} else {
	    recognize.setEnabled( false );
	    obey.setEnabled( false );
	}
	// Set the window title
	setTitle( WINDOW_TITLE_STEM + sset.getName() );
	// Regenerate the Lexicion sub-menu
	buildLexiconMenu();
    }

    // }}} loadScenarioSet()

    // {{{ saveScenarioSet()

    /**
     * Saves a ScenarioSet to the given File.  If the file already
     * exists, the user is queried as to whether or not they want to
     * replace that file.  If the file cannot be created, the user it
     * notified.  If the file cannot be written, the user is notified.
     *
     * @param     file The File object corresponding to the file to
     *            save the ScenarioSet to. 
     **/
    protected void saveScenarioSet(File file)
    {
	// Simple check
	if (file == null) {
	    JOptionPane.showMessageDialog(this,
					  "No filename specified.",
					  "No filename",
					  JOptionPane.ERROR_MESSAGE);
	}

	if ( file.exists() ) {
	    if (JOptionPane.showConfirmDialog(this,
					      file.getAbsolutePath() +
					      " exists.  Replace?",
					      "File replacement confirmation",
					      JOptionPane.YES_NO_OPTION,
					      JOptionPane.WARNING_MESSAGE)
		== JOptionPane.NO_OPTION) {
		// They don't want to replace the file, so we stop here.
		return;
	    }
	}
	
	try {
	    // FIXME: Write something similar to
	    // ProgressMonitorInputStream
	    // and use a
	    // com.sun.java.swing.ProgressMonitor 
	    FileOutputStream fos = new FileOutputStream( file );
	    ObjectOutputStream oos = new ObjectOutputStream( fos );
	    System.out.println("About to write out ScenarioSet to " +
			       file.getAbsolutePath() );
	    oos.writeObject( sset );
	    System.out.println("Wrote out ScenarioSet");
	} catch (FileNotFoundException fnfe) {
	    // Shouldn't happen, but might.
	    JOptionPane.showMessageDialog(this,
					  "Cannot find file " +
					  file.getAbsolutePath(),
					  "Cannot find file.",
					  JOptionPane.ERROR_MESSAGE);
	} catch (IOException ioe) {
	    Object msgs[] = new String[2];
	    msgs[0] = new String("Cannot write to file " +
				 file.getAbsolutePath() );
	    msgs[1] = new String("IOException message: " + ioe.getMessage());
	    JOptionPane.showMessageDialog(this,
					  msgs,
					  "Cannot save file.",
					  JOptionPane.ERROR_MESSAGE);

	} // end try-catch-catch
    }

    // }}} saveScenarioSet()

    // {{{ buildLexiconMenu()

    /**
     * Clears and builds the menu
     * <A HREF="#openLexiconMenu">openLexiconMenu</A> which allows
     * the opening of Lexicon viewers.
     **/
    protected void buildLexiconMenu()
    {
	// Clear the menu.
	openLexiconMenu.removeAll();
	// Add menu items for the ScenarioSet's available Lexicons
	if (sset != null) {
	    Vector lexicons = sset.getLexicons();
	    Lexicon lex;

	    for (int i=0 ; i <lexicons.size(); i++) {
		lex = (Lexicon) lexicons.elementAt(i);
		openLexiconMenu.add( new OpenLexicon(lex) );
	    } // end for over i
	}
    }

    // }}} buildLexiconMenu()

    // {{{ public static void main(...)
    public static void main(String args[])
    {
	// Load the properties...
	if (vlProps == null) {
	    try {
		vlProps =
		    NTLLauncher.loadNTLSystemProperties("verblearn");
	    } catch (FileNotFoundException fnfe) {
		System.out.println("Properties file not found: " +
				   fnfe.getMessage() );
		fnfe.printStackTrace();
	    } catch (IOException ioe) {
		System.out.println("I/O error while reading properties file: " +
				   ioe.getMessage() );
		ioe.printStackTrace();
	    }
	}

	// Load the Feature Dictionary
	String fdFN =
	    vlProps.getProperty("ntlsys.ntlhome") + "/" +
	    vlProps.getProperty("ntlsys.verblearn.homeDir") +
	    vlProps.getProperty("ntlsys.verblearn.dataDir") +
	    vlProps.getProperty("ntlsys.verblearn.dictDir") +
	    vlProps.getProperty("ntlsys.verblearn.dictFN");
	try {
	    featureDict = FeatureDictionary.loadDict( new File( fdFN ) );
	} catch (FileNotFoundException fnfe) {
	    System.out.println("Could not find Dictionary " + fdFN );
	    System.out.println("Using new empty FeatureDictionary.");
	} catch (ClassNotFoundException cnfe) {
	    System.out.println( cnfe.getMessage() );
	    cnfe.printStackTrace();
	} catch (IOException ioe) {
	    System.out.println( ioe.getMessage() );
	    ioe.printStackTrace();
	} // end try-catch

	// Now that the Properties and FeatureDictionary are
	// loaded, we can create a VerbLearn instance.
	VerbLearn vl = new VerbLearn();
	WindowWaiter ww = new WindowWaiter();
	vl.addWindowListener(ww);
	ww.WaitUntilDone();

	// Save the FeatureDictionary.
	try {
	    ObjectOutputStream oos =
		new ObjectOutputStream(new FileOutputStream( fdFN ));
	    oos.writeObject( featureDict );
	    oos.close();
	} catch (FileNotFoundException fnfe) {
	    System.out.println( fnfe.getMessage() );
	    fnfe.printStackTrace();
	} catch (IOException ioe) {
	    System.out.println( ioe.getMessage() );
	    ioe.printStackTrace();
	} // end try-catch

	System.exit(-1);

    }
    // }}}

    /**
     * Returns the currently-being-edited ScenarioSet.  This method
     * may return null if no ScenarioSet is currently being edited.
     *
     * @return    The currently-loaded ScenarioSet, or null if no
     *            ScenarioSet is loaded.
     **/
    public ScenarioSet getScenarioSet()
    {
	return sset;
    }

    // {{{ FeatureDictionary accessors
    
    public static FeatureDictionary getFeatureDict()
    { return featureDict; }

    // }}} FeatureDictionary accessors

    
    // }}} Methods

    // {{{ Inner classes

    /**
     * An AbstractAction which easily opens a LexiconViewer for a
     * given Lexicon object.
     **/
    protected class OpenLexicon
	extends AbstractAction
    {
	protected Lexicon toOpen;

	/**
	 * Constructor for the OpenLexicon Action.  Sets the Action
	 * name to be that of the lexicon name, and stores a pointer
	 * to the given Lexicon lex.
	 *
	 * @param     lex The Lexicon to open in the viewer
	 **/
	public OpenLexicon( Lexicon lex  )
	{
	    super( lex.getName() );
	    toOpen = lex;
	}

	public void actionPerformed(ActionEvent ae)
	{
	    LexiconViewer lexv =
		new LexiconViewer( VerbLearn.this, toOpen );
	    desktop.add( lexv,
			 JLayeredPane.DEFAULT_LAYER );
	    desktop.moveToFront( lexv );
	}
    }
    // }}} Inner classes
}
