package NTLSystem.verblearn;

import NTLSystem.features.*;

/**
 * An extension of DiscreteDist which handles a seperate virtual
 * sample count.  This class was created specifically for use with
 * VerbLearn.
 * <P>
 * Based on dbailey's old VerbLearn: Lexicon.java::Dist class.
 * <P>
 * $Id: VirtualSampleDiscreteDist.java,v 1.5 1998/10/16 01:59:44 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Sense
 * @version   $Revision: 1.5 $
 **/
public class VirtualSampleDiscreteDist
    extends DiscreteDist
{

    /**
     * Serialization stream unique identifier.  Recompute [with
     * serialver] if this class changes.
     **/
    static final long serialVersionUID = 8864702656264337080L;

    /**
     * The total number of virtual samples [ size * virt ] in use.
     **/
    private double kVirt;

    /**
     * The number of virtual samples.
     **/
    private double virt;

    /**
     * Basic no-arg constructor for serialization.
     **/
    public VirtualSampleDiscreteDist()
    { super(); }

    /**
     * Constructs a VirtualSampleDiscreteDist with the given values
     * and the given virtualSample count.  This is probably the most
     * useful constructor.
     *
     * @param     values The initial values for the DiscreteDist.
     * @param     virtualSamples The virtual sample count for the
     *            VirtualSampleDiscreteDist.
     **/
    public VirtualSampleDiscreteDist( double[] values,
				      double virtualSamples )
    {
	super(values);
	virt = virtualSamples;
	kVirt = values.length * virtualSamples;
    }

    /**
     * Creates an empty VirtualSampleDiscreteDist with the given
     * number of possible values and virtual sample count.
     *
     * @param     size The number of discrete elements in this VSDD.
     * @param     vSamps The virtual sample count.
     **/
    public VirtualSampleDiscreteDist(int size, double vSamps)
    {
	super( size );
	virt = vSamps;
	kVirt = size * vSamps;
    }

    /**
     * Merges two VirtualSampleDiscreteDists together.  The value
     * counts are preserved, and the new virtual sample value is
     * given. 
     *
     * @param     d1 The first VirtualSampleDiscreteDist to merge.
     * @param     d2 The second VirtualSampleDiscreteDist to merge.
     * @param     newVirtual The new virtual sample count to use.
     * @exception IllegalArgumentException If the dists are of
     *            different sizes.
     **/
    public VirtualSampleDiscreteDist( VirtualSampleDiscreteDist d1,
				      VirtualSampleDiscreteDist d2,
				      double newVirtual )
	throws IllegalArgumentException
    {
	super( d1.getSize() );
	if ( d1.getSize() != d2.getSize() ) {
	    // FIXME: that's a problem.  They should be the same.
	    throw new IllegalArgumentException( "Dists are of different sizes (" +
						d1.getSize() + ", " +
						d2.getSize() + ")." );
	}
	virt = newVirtual;
	kVirt = getSize() * virt;
	for (int i=0; i<d1.getSize(); i++) {
	    setProb( i, d1.getCount(i) + d2.getCount(i) );
	} // end for over i
	setSum( d1.getSum() + d2.getSum() );
    }

    /**
     * Creates a new VirtualSampleDiscreteDist with the same values as
     * the given DiscreteDist, and with the given virtual sample
     * value. 
     *
     * @param     dd The DiscreteDist to base this
     *            VirtualSampleDiscreteDist off of.
     * @param     virt The virtual sample count to use.
     **/
    public VirtualSampleDiscreteDist(DiscreteDist dd,
				     double virt)
    {
	double values[] = new double[ dd.getSize() ];
	double sum = 0.0;
	for (int i=0; i < values.length; i++) {
	    values[i] = dd.getProb( i );
	    sum += values[i];
	} // end for over i
	this.values = values;
	this.sum = sum;
	this.virt = virt;
	kVirt = getSize() * virt;
    }

    /**
     * Returns the number of virtual samples in this
     * VirtualSampleDiscreteDist.
     *
     * @return     The virtual sample count for this dist.
     **/
    public double getVirtualSamples()
    { return virt; }

    /**
     * Sets the virtual sample count for this dist.
     *
     * @param     newVirt The new virtual sample count to use for this
     *            dist.
     **/
    public void setVirtualSamples(double newVirt)
    {
	virt = newVirt;
	kVirt = getSize() * virt;
    }

    /**
     * Make DiscreteDist.normalize a no-op, just in case it's called.
     * We need to do this because VirtualSampleDiscreteDists don't 
     * really have probabilities of values so much as they have
     * 'counts' of observed values.  Normalization destroys these
     * counts, and therefore must be avoided.  getProb, however, will
     * report the correct value as it divides by 'sum'.
     **/
    public void normalize()
    {
	; // no-op
    }

    /**
     * A modified version of DiscreteDist.getProb(int) which correctly
     * handles the virtual samples.
     *
     * @param     i The index of the discrete value to get.
     * @return    The correct likelihood for this
     *            VirtualSampleDiscreteDist.
     **/
    public double getProb(int i)
    {
	double toRet = 0.0;
	if ( (sum + kVirt) != 0.0 ) {
	    return (double)(getCount(i) + virt) / (double)(sum + kVirt);
	}
	return toRet;
    }


    /**
     * A heuristic measure of the difference between two
     * distributions.  Basically chi-square, except we use a
     * normalized version to factor out differences in number of
     * samples, and then invert it with 1/(1+chi) to give a measure
     * ranging from 0 (very different) to 1 (identical).
     * <P>
     * See <A HREF="#normalizedChiSquaredWith(DiscreteDist)">
     * normalizedChiSquareWidth( DiscreteDist )</A>
     * for more info.
     *
     * @param     dd Another VirtualSampleDiscreteDist to compute our
     *            similarity to.
     * @return    The similarity to the given DiscreteDist.
     * @see       #normalizedChiSquareWith(DiscreteDist)
     * @see       NTLSystem.verblearn.Sense
     **/
    public double similarityTo(VirtualSampleDiscreteDist dd)
    {
	return 1 / (1 + normalizedChiSquareWith( dd ));
    }


    /**
     * From Numerical Recipes 2nd ed., p. 623 ("r" and "s" in that
     * alg. map to "this" and "d" respectively).
     * <P>
     * Probabilities rather than counts are used, since otherwise chi
     * square is problematic because it also measures differences in
     * the number of samples in the two distributions.
     * <P>
     * <STRONG>   NOTE:   </STRONG>
     * The two DiscreteDists being compared should have the same
     * FeatureDescriptors [and thus the same discrete values in the
     * same indices], or the value returned from this method will not
     * make sense.
     *
     * @param     d Another VirtualSampleDiscreteDist to compute the
     *            normalized chi-squared value with respect to.
     * @return    The normalized chi-squared value w.r.t. 'd'.
     * @see       NTLSystem.verblearn.Sense
     **/
    protected double normalizedChiSquareWith(VirtualSampleDiscreteDist d) {
	double res = 0.0;
	for (int i=0; i<values.length; i++) {
	    double ri = getProb( i );
	    double si = d.getProb( i );
	    if (ri != 0.0 || si != 0.0) {
		res += Math.pow(ri - si, 2) / (ri + si);
	    }
	}
	return res;
    }

    /**
     * A special method which returns the prob for a value as
     * DiscreteDist would return it.  This is useful for merging two
     * VirtualSampleDiscreteDists, as their 'count's need to be
     * combined, without the influence of virtual samples.
     *
     * @param     i The index to return the count for.
     * @return    The original probability for the given value before
     *            being influenced by virtual samples.
     **/
    public double getCount(int i)
    {
	return super.getProb(i);

    }

    /**
     * Redefinition of clone
     **/
    public Object clone()
    {
	VirtualSampleDiscreteDist toRet =
	    (VirtualSampleDiscreteDist)super.clone();
	toRet.virt = this.virt;
	toRet.kVirt = this.kVirt;
	return toRet;
    }

    public String toString()
    {
	StringBuffer toRet = new StringBuffer(100);
	toRet.append(" VSDD { ");
	toRet.append(" virt: ").append(virt);
	toRet.append(", kVirt: " ).append(kVirt);
	toRet.append("; ");
	for (int i=0; i<values.length-1; i++) {
	    toRet.append( getProb(i) ).append(" (").
		append(values[i]).append(")").append(", ");
	}
	if (values.length-1 >= 0) {
	    toRet.append( getProb(values.length-1) ).append(" (").
		append(values[values.length-1]).append(")");
	}
	toRet.append(" }");
	return toRet.toString();
    }


}
